from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui
from numpy import inf
import re

"""
This code has been adapted from: https://gist.github.com/jdreaver/0be2e44981159d0854f5
Changes made are support for PyQt5, localisation, better intermediate state detection and better stepping.
Some inspiration taken from: https://github.com/pyqtgraph/pyqtgraph/blob/develop/pyqtgraph/widgets/SpinBox.py
"""

decimal_point = re.escape(QtCore.QLocale().decimalPoint())
exp_regex = r'(([+-]?\d+(' + decimal_point + r'\d*)?|' + decimal_point + r'\d+)([eE][+-]?\d+)?)'
exp_float_re = re.compile(exp_regex)
partial_int_re = re.compile(r'([+-]?(\d?))')
partial_float_regex = r'([+-]?((\d*' + decimal_point + r'?))?\d*)'
partial_float_re = re.compile(partial_float_regex)


class FloatValidator(QtGui.QValidator):
	"""
	Validates float inputs for ScientificDoubleSpinBox
	"""

	@staticmethod
	def valid_float_string(string: str) -> bool:
		"""
		Checks if string makes a valid float, keeping in mind locale dependent decimal separators.
		"""

		match = exp_float_re.search(string)
		return match.groups()[0] == string if match else False

	@staticmethod
	def intermediate_integer_string(string: str) -> bool:
		"""
		Checks if string makes a valid partial integer, keeping in mind locale dependent decimal separators.
		"""

		match = partial_int_re.search(string)
		return match.groups()[0] == string if match else False

	@staticmethod
	def intermediate_float_string(string: str) -> bool:
		"""
		Checks if string makes a valid partial float, keeping in mind locale dependent decimal separators.
		"""

		# Allow empty strings:
		if string == "":
			return True

		# Allow 0 or 1 combined instances of e and E:
		exp_count = string.count('e') + string.count('E')
		if exp_count > 1:
			return False

		if exp_count == 1:
			# Scientific notation.
			# Split string at exponential symbol and check that both substrings are valid:
			char = 'e' if string.count('e') > 0 else 'E'
			_, __ = string.split(char)
			if FloatValidator.intermediate_float_string(_) and FloatValidator.intermediate_integer_string(__):
				return True
		else:
			# Normal notation.
			# Use regex to check if string is part of a valid float:
			match = partial_float_re.search(string)
			return match.groups()[0] == string if match else False

		return False

	def validate(self, string: str, position: int) -> QtGui.QValidator.State:
		"""
		Validates input string to see if it is a valid float or partial float.

		Notes
		-----
		State is not used, but required because we are overriding an internal method.
		"""

		if self.valid_float_string(string):
			return QtGui.QValidator.Acceptable
		if self.intermediate_float_string(string):
			return QtGui.QValidator.Intermediate
		else:
			return QtGui.QValidator.Invalid

	def fixup(self, string: str) -> str:
		"""
		Fixes up input text to create a valid float. Puts an empty string on failure.
		"""

		match = exp_float_re.search(string)
		return match.groups()[0] if match else ""


class ScientificDoubleSpinBox(QtWidgets.QDoubleSpinBox):
	"""
	Subclass of QDoubleSpinBox that allows for scientific notation and is locale independent.
	"""

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		# Set nice default values:
		self.setMinimum(-inf)
		self.setMaximum(inf)
		self.validator = FloatValidator()
		self.setDecimals(1000)

	def validate(self, string: str, position: int) -> (QtGui.QValidator.State, str, int):
		"""
		Returns the validity of the string, using a QtGui.QValidator object.

		Notes
		-----
		Return type depends on whether PySide or PyQt is being used.
		"""

		validity = self.validator.validate(string, position)

		# support 2 different PyQt APIs.
		if hasattr(QtCore, 'QString'):
			return validity, position
		else:
			return validity, string, position

	def fixup(self, string: str) -> str:
		"""
		Fixes up strings that are considered intermediate when spinbox focus is lost.
		"""

		return self.validator.fixup(string)

	def textFromValue(self, value: float) -> str:
		"""
		Turns internal float value into user viewable string.
		"""

		string = "{:g}".format(value).replace("e+", "e").replace(".", QtCore.QLocale().decimalPoint(), 1)
		string = re.sub("e(-?)0*(\d+)", r"e\1\2", string)
		return string

	def valueFromText(self, string: str) -> float:
		"""
		Turns user viewable string into a float.
		"""

		if string == "":
			return self.value()

		string = string.replace(QtCore.QLocale().decimalPoint(), ".", 2)
		value = float(string)

		if value > self.maximum():
			value = self.maximum()
		if value < self.minimum():
			value = self.minimum()

		return value

	def stepBy(self, steps: int):
		"""
		Increases or decreases the value of the least significant digit by 1.
		"""

		text = self.cleanText()

		groups = exp_float_re.search(text).groups()

		# If we have digits after the decimal indicator:
		if groups[2] is not None:
			significance = len(groups[2])-1
			if len(groups[2]) > 0 and groups[2][0] == '-':
				significance -= 1
			val = float(groups[1]) + steps * 10**(-significance)

		else:
			val, _, __ = groups[1].partition(QtCore.QLocale().decimalPoint())
			significance = len(groups[1])-1
			if groups[1] is not None and len(groups[1]) > 0 and groups[1][0] == '-':
				significance -= 1
			val = float(val) + steps * 10**significance

		new_string = "{:g}".format(val) + (groups[3] if groups[3] else "")
		self.lineEdit().setText(new_string)



from functools import reduce
from copy import deepcopy
import numpy as np
import benedict


class DictValueNotEqualException(Exception):
    """
    Exception used to break out of the dictionary comparison loop immediately if two values do not match
    """
    pass


def compare_dicts_floating_precision(dict_1: dict, dict_2: dict, atol: float) -> bool:
    """
    Compares two dictionaries recursively. Early return ``False`` if any nested dictionary is found to have a different
    length, and keys are found not to match, any float values found to not match by the floating-point precision
    specified by ``atol``, or any non-float values found not to match.

    Parameters
    ----------
    dict_1: dict
        First dictionary to compare

    dict_2: dict
        Second dictionary to compare

    atol: float
        Floating-point precision used to compare float values

    Returns
    -------
    bool
        ``True`` if the dictionaries match, ``False`` otherwise
    """
    def compare(d1: dict, d2: dict):
        for (k1, v1), (k2, v2) in zip(d1.items(), d2.items()):
            if k1 != k2:
                # Found non-matching keys
                raise DictValueNotEqualException
            if isinstance(v1, dict) and isinstance(v2, dict):
                if len(v1) != len(v2):
                    # Found dictionaries of different length
                    raise DictValueNotEqualException
                compare(v1, v2)
            else:
                if isinstance(v1, float) and isinstance(v2, float):
                    if abs(v1 - v2) > atol:
                        # Found float values not equal
                        raise DictValueNotEqualException
                else:
                    if v1 != v2:
                        # Found non-float values not equal
                        raise DictValueNotEqualException

    try:
        compare(dict_1, dict_2)
        return True
    except DictValueNotEqualException:
        return False


def set_all_dict_values(d: dict):
    for k, v in d.items():
        if isinstance(v, dict):
            set_all_dict_values(v)
        else:
            v.param_dict = d


def assign_airfoil_tags_to_param_dict(d: dict, airfoil_tag: str):
    for k, v in d.items():
        if isinstance(v, dict):
            assign_airfoil_tags_to_param_dict(v, airfoil_tag)
        else:
            v.airfoil_tag = airfoil_tag


def assign_names_to_params_in_param_dict(d: dict):
    dben = benedict.benedict(d)
    keypaths = dben.keypaths()
    for k in keypaths:
        p = dben[k]
        if not isinstance(p, dict):
            p.name = k


def recursive_get(d, *keys):
    """From answer by Thomas Orozco (https://stackoverflow.com/a/28225747)"""
    return reduce(lambda c, k: c.get(k, {}), keys, d)


def unravel_param_dict(d: dict, output_dict: dict, prep_for_json: bool = True):
    for k, v in d.items():
        if isinstance(v, dict):
            output_dict[k] = {}
            unravel_param_dict(v, output_dict[k], prep_for_json=prep_for_json)
        else:
            param_dict_attrs = vars(v)
            for attr_key, attr in param_dict_attrs.items():
                if prep_for_json:
                    if isinstance(attr, np.ndarray):
                        param_dict_attrs[attr_key] = attr.tolist()
                    if attr_key in ['mea', 'param_dict', 'free_point', 'anchor_point', 'affects', 'depends_on',
                                    'function_dict', 'func']:
                        param_dict_attrs[attr_key] = None
            output_dict[k] = param_dict_attrs


def unravel_param_dict_deepcopy(d: dict, output_dict: dict):
    for k, v in d.items():
        if isinstance(v, dict):
            output_dict[k] = {}
            unravel_param_dict_deepcopy(v, output_dict[k])
        else:
            param_dict_attrs = vars(v)
            temp_dict = {}
            for attr_key, attr in param_dict_attrs.items():
                if attr_key in ['name', '_value', 'active', 'linked', 'func_str', 'bounds']:
                    if isinstance(attr, np.ndarray):
                        temp_dict[attr_key] = deepcopy(attr.tolist())
                    else:
                        temp_dict[attr_key] = deepcopy(attr)
            output_dict[k] = temp_dict



import os

from PyQt5.QtCore import Qt, QSize, QUrl, pyqtSignal
from PyQt5.QtGui import QIcon, QDesktopServices, QPalette, QLinearGradient, QGradient, QBrush
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QLabel, QToolButton, QStyle

from pymead import ICON_DIR


class TitleBarButton(QToolButton):
    def __init__(self, parent, operation: str, theme: dict = None):
        super().__init__(parent)

        if operation not in ("minimize", "maximize", "normal", "close"):
            raise ValueError("Invalid operation")

        self.operation = operation
        self.theme = theme

    def hoverColor(self) -> str:
        if self.theme is not None:
            return self.theme[f"{self.operation}-hover-color"]
        else:
            return self.parent().parent().themes[self.parent().parent().current_theme][f"{self.operation}-hover-color"]

    def setColorDefault(self):
        theme_name = self.theme["theme-name"] if self.theme is not None else self.parent().parent().current_theme
        self.setIcon(QIcon(os.path.join(ICON_DIR, f"{self.operation}-{theme_name}-mode.svg")))
        self.setStyleSheet(f"""QToolButton {{ border: none }}""")

    def setColorHover(self):
        theme_name = self.theme["theme-name"] if self.theme is not None else self.parent().parent().current_theme
        if theme_name == "dark":
            self.setIcon(QIcon(os.path.join(ICON_DIR, f"{self.operation}-{theme_name}-mode.svg")))
        self.setStyleSheet(f"""QToolButton {{ background-color: {self.hoverColor()} }}""")

    def enterEvent(self, a0):
        self.setColorHover()
        super().enterEvent(a0)

    def leaveEvent(self, a0):
        self.setColorDefault()
        super().leaveEvent(a0)


class TitleBar(QWidget):
    clickPos = None

    sigMessage = pyqtSignal(str, str)

    def __init__(self, parent):
        super().__init__(parent)

        self.guiWindowMaximized = False

        theme = self.parent().themes[self.parent().current_theme]

        self.setAutoFillBackground(True)

        self.lay = QHBoxLayout(self)
        self.lay.setContentsMargins(1, 1, 1, 1)

        self.title = QLabel("Custom Title Bar", self)
        self.title.setAlignment(Qt.AlignCenter)
        self.title.setFixedHeight(30)
        self.title.setAutoFillBackground(True)

        # style = self.style()
        # ref_size = self.fontMetrics().height()
        # ref_size += style.pixelMetric(style.PM_ButtonMargin) * 2
        # self.setMaximumHeight(ref_size + 2)

        # Add the pymead logo as a button that, when clicked, opens the pymead GitHub page in the machine's default
        # browser
        pymeadLogoButton = QToolButton(self)
        pymeadLogoButton.setIcon(QIcon(os.path.join(ICON_DIR, "pymead-logo.png")))
        pymeadLogoButton.setFixedSize(30, 30)
        pymeadLogoButton.setIconSize(QSize(30, 30))
        pymeadLogoButton.setToolTip("Open pymead GitHub repo")
        pymeadLogoButton.clicked.connect(self.openGitHubPage)

        gripFillerWidget = QWidget(self)
        gripFillerWidget.setFixedWidth(0)

        self.lay.addWidget(gripFillerWidget)
        self.lay.addWidget(pymeadLogoButton)
        self.lay.addWidget(self.title)

        # self.lay.addStretch()

        self.minimizeButton = None
        self.normalButton = None
        self.maximizeButton = None
        self.closeButton = None

        btn_size = QSize(30, 30)
        for target, picture, hover_color, tool_tip in zip(
                ('minimize', 'normal', 'maximize', 'close'),
                ("minimize-dark-mode.svg", "normal-dark-mode.svg", "maximize-dark-mode.svg", "close-dark-mode.svg"),
                (theme["minimize-hover-color"], theme["maximize-hover-color"],
                 theme["maximize-hover-color"], theme["close-hover-color"]),
                ("Minimize", "Normal", "Maximize", "Close")
        ):
            btn = TitleBarButton(self, operation=target)
            btn.setFocusPolicy(Qt.NoFocus)
            self.lay.addWidget(btn)
            btn.setFixedSize(btn_size)
            # btn.setIconSize(btn_size)

            btn.setIcon(QIcon(os.path.join(ICON_DIR, picture)))
            btn.setToolTip(tool_tip)

            btn.setStyleSheet('''QToolButton { border: none }''')

            signal = getattr(self, target + 'Clicked')
            btn.clicked.connect(signal)

            setattr(self, target + 'Button', btn)

        gripFillerWidget = QWidget(self)
        gripFillerWidget.setFixedWidth(0)

        self.lay.addWidget(gripFillerWidget)

        self.normalButton.hide()

        self.updateTitle(parent.windowTitle())
        parent.windowTitleChanged.connect(self.updateTitle)

    def updateTitle(self, title=None):
        if title is None:
            title = self.window().windowTitle()
        width = self.title.width()
        width -= self.style().pixelMetric(QStyle.PM_LayoutHorizontalSpacing) * 2
        self.title.setText(self.fontMetrics().elidedText(
            title, Qt.ElideRight, width))
        # self.title.setStyleSheet("QLabel { color: blue; }")

    def windowStateChanged(self, state):
        self.normalButton.setVisible(state == Qt.WindowMaximized)
        self.maximizeButton.setVisible(state != Qt.WindowMaximized)
        self.guiWindowMaximized = state == Qt.WindowMaximized

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and not self.guiWindowMaximized:
            self.clickPos = event.windowPos().toPoint()
        super().mousePressEvent(event)
        event.accept()

    def mouseMoveEvent(self, event):
        if self.clickPos is not None and not self.guiWindowMaximized:
            self.window().move(event.globalPos() - self.clickPos)
        super().mouseMoveEvent(event)
        event.accept()

    def mouseReleaseEvent(self, event):
        self.clickPos = None
        super().mouseReleaseEvent(event)
        event.accept()

    def closeClicked(self):
        self.window().close()

    def maximizeClicked(self):
        self.window().showMaximized()
        self.maximizeButton.setColorDefault()

    def normalClicked(self):
        self.window().showNormal()
        self.normalButton.setColorDefault()

    def minimizeClicked(self):
        self.window().showMinimized()

    def openGitHubPage(self):
        url = QUrl("https://github.com/mlau154/pymead")
        if not QDesktopServices.openUrl(url):
            self.sigMessage.emit("Could not open pymead's GitHub page", "error")


class DialogTitleBar(QWidget):
    clickPos = None

    sigMessage = pyqtSignal(str, str)

    def __init__(self, parent, theme: dict):
        super().__init__(parent)

        self.guiWindowMaximized = False

        self.setAutoFillBackground(True)

        self.lay = QHBoxLayout(self)
        self.lay.setContentsMargins(1, 1, 1, 1)

        self.title = QLabel("Custom Title Bar", self)
        self.title.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)
        self.title.setFixedHeight(30)
        self.title.setMinimumWidth(400)
        self.title.setAutoFillBackground(True)

        # style = self.style()
        # ref_size = self.fontMetrics().height()
        # ref_size += style.pixelMetric(style.PM_ButtonMargin) * 2
        # self.setMaximumHeight(ref_size + 2)

        # Add the pymead logo as a button that, when clicked, opens the pymead GitHub page in the machine's default
        # browser
        pymeadLogoButton = QToolButton(self)
        pymeadLogoButton.setIcon(QIcon(os.path.join(ICON_DIR, "pymead-logo.png")))
        pymeadLogoButton.setFixedSize(30, 30)
        pymeadLogoButton.setIconSize(QSize(30, 30))
        # pymeadLogoButton.setEnabled(False)

        gripFillerWidget = QWidget(self)
        gripFillerWidget.setFixedWidth(0)

        self.lay.addWidget(gripFillerWidget)
        self.lay.addWidget(pymeadLogoButton)
        self.lay.addWidget(self.title)

        # self.lay.addStretch()

        self.minimizeButton = None
        self.normalButton = None
        self.maximizeButton = None
        self.closeButton = None

        btn_size = QSize(30, 30)
        for target, picture, hover_color, tool_tip in zip(
                ('close',),
                (f"close-{theme['theme-name']}-mode.svg",),
                (theme["close-hover-color"],),
                ("Close",)
        ):
            btn = TitleBarButton(self, operation=target, theme=theme)
            btn.setFocusPolicy(Qt.NoFocus)
            self.lay.addWidget(btn)
            btn.setFixedSize(btn_size)
            # btn.setIconSize(btn_size)

            btn.setIcon(QIcon(os.path.join(ICON_DIR, picture)))
            btn.setToolTip(tool_tip)

            btn.setStyleSheet('''QToolButton { border: none }''')

            signal = getattr(self, target + 'Clicked')
            btn.clicked.connect(signal)

            setattr(self, target + 'Button', btn)

        gripFillerWidget = QWidget(self)
        gripFillerWidget.setFixedWidth(0)

        self.lay.addWidget(gripFillerWidget)

        self.updateTitle(parent.windowTitle())
        parent.windowTitleChanged.connect(self.updateTitle)

    def updateTitle(self, title=None):
        if title is None:
            title = self.window().windowTitle()
        width = self.title.width()
        width -= self.style().pixelMetric(QStyle.PM_LayoutHorizontalSpacing) * 2
        self.title.setText(self.fontMetrics().elidedText(
            title, Qt.ElideRight, width))
        # self.title.setStyleSheet("QLabel { color: blue; }")

    def windowStateChanged(self, state):
        self.normalButton.setVisible(state == Qt.WindowMaximized)
        self.maximizeButton.setVisible(state != Qt.WindowMaximized)
        self.guiWindowMaximized = state == Qt.WindowMaximized

    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton and not self.guiWindowMaximized:
            self.clickPos = event.windowPos().toPoint()
        super().mousePressEvent(event)
        event.accept()

    def mouseMoveEvent(self, event):
        if self.clickPos is not None and not self.guiWindowMaximized:
            self.window().move(event.globalPos() - self.clickPos)
        super().mouseMoveEvent(event)
        event.accept()

    def mouseReleaseEvent(self, event):
        self.clickPos = None
        super().mouseReleaseEvent(event)
        event.accept()

    def closeClicked(self):
        self.window().close()

    def openGitHubPage(self):
        url = QUrl("https://github.com/mlau154/pymead")
        if not QDesktopServices.openUrl(url):
            self.sigMessage.emit("Could not open pymead's GitHub page", "error")



from PyQt5.QtWidgets import QWidget


def get_parent(widget: QWidget, depth: int = 1):
    depth -= 1
    if depth == 0:
        return widget.parent()
    else:
        if widget.parent():
            return get_parent(widget.parent(), depth=depth)
        else:
            return None



import os

from PyQt5.QtWidgets import QToolBar, QAction, QLineEdit, QStatusBar, QWidget, QVBoxLayout, QToolButton, QDialog
from PyQt5.QtWebEngineWidgets import QWebEngineView
from PyQt5.QtCore import QUrl
from PyQt5.QtGui import QIcon

from pymead import ICON_DIR


class HelpBrowserToolBar(QToolBar):
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.back_icon = QIcon(os.path.join(ICON_DIR, "back.png"))
        self.forward_icon = QIcon(os.path.join(ICON_DIR, "forward.png"))
        self.reload_icon = QIcon(os.path.join(ICON_DIR, "reload.png"))
        self.home_icon = QIcon(os.path.join(ICON_DIR, "home.png"))
        self.stop_icon = QIcon(os.path.join(ICON_DIR, "stop.png"))
        self.back_button = QToolButton(parent=self)
        self.forward_button = QToolButton(parent=self)
        self.reload_button = QToolButton(parent=self)
        self.home_button = QToolButton(parent=self)
        self.url_bar = QLineEdit(parent=self)
        self.stop_button = QToolButton(parent=self)
        self.back_button.setIcon(self.back_icon)
        self.forward_button.setIcon(self.forward_icon)
        self.reload_button.setIcon(self.reload_icon)
        self.home_button.setIcon(self.home_icon)
        self.stop_button.setIcon(self.stop_icon)
        for w in [self.back_button, self.forward_button, self.reload_button, self.home_button, self.url_bar,
                  self.stop_button]:
            self.addWidget(w)


class HelpBrowser(QWebEngineView):
    """
    Help browser as inspired by https://www.geeksforgeeks.org/creating-a-simple-browser-using-pyqt5/
    """
    def __init__(self, parent):
        super().__init__(parent=parent)
        self.setUrl(QUrl("https://pymead.readthedocs.io/en/latest/gui.html"))


class HelpBrowserWidget(QWidget):
    def __init__(self, parent):
        super().__init__(parent=parent)

        self.lay = QVBoxLayout()
        self.setLayout(self.lay)
        self.toolbar = HelpBrowserToolBar(self)
        self.lay.addWidget(self.toolbar)
        self.help_browser = HelpBrowser(self)
        self.lay.addWidget(self.help_browser)
        self.help_browser.urlChanged.connect(self.update_urlbar)

        # Connections
        self.toolbar.back_button.clicked.connect(self.help_browser.back)
        self.toolbar.forward_button.clicked.connect(self.help_browser.forward)
        self.toolbar.reload_button.clicked.connect(self.help_browser.reload)
        self.toolbar.home_button.clicked.connect(self.navigate_home)
        self.toolbar.url_bar.returnPressed.connect(self.navigate_to_url)
        self.toolbar.stop_button.clicked.connect(self.help_browser.stop)

    # method called by the home action
    def navigate_home(self):

        # open the google
        self.help_browser.setUrl(QUrl("https://pymead.readthedocs.io/en/latest/gui.html"))

    # method called by the line edit when return key is pressed
    def navigate_to_url(self):

        # getting url and converting it to QUrl object
        q = QUrl(self.toolbar.url_bar.text())

        # if url is scheme is blank
        if q.scheme() == "":
            # set url scheme to html
            q.setScheme("http")

        # set the url to the browser
        self.help_browser.setUrl(q)

    # method for updating url
    # this method is called by the QWebEngineView object
    def update_urlbar(self, q):

        # setting text to the url bar
        self.toolbar.url_bar.setText(q.toString())

        # setting cursor position of the url bar
        self.toolbar.url_bar.setCursorPosition(0)


class HelpBrowserWindow(QDialog):
    def __init__(self, parent):
        super().__init__(parent)

        self.setWindowTitle("Help Browser")
        if self.parent() is not None:
            self.setFont(self.parent().font())
        self.setGeometry(100, 50, 800, 800)

        self.help_browser_widget = HelpBrowserWidget(self)
        self.lay = QVBoxLayout()
        self.setLayout(self.lay)
        self.lay.addWidget(self.help_browser_widget)
        self.show()



from PyQt5.QtWidgets import QLineEdit, QCompleter
from PyQt5.QtCore import Qt
from pyqtgraph.parametertree.parameterTypes.basetypes import WidgetParameterItem

from pymead.gui.custom_context_menu_event import custom_context_menu_event


class AutoStrParameterItem(WidgetParameterItem):
    """Parameter type which displays a QLineEdit with an auto-completion mechanism built in"""

    def __init__(self, param, depth):
        self.widget = None
        super().__init__(param, depth)

    def makeWidget(self):
        w = QLineEdit()
        completer = Completer()
        w.setCompleter(completer)
        w.setStyleSheet('border: 0px')
        w.sigChanged = w.editingFinished
        w.value = w.text
        w.setValue = w.setText
        w.sigChanging = w.textChanged
        w.setPlaceholderText("$")
        self.widget = w
        return w

    def contextMenuEvent(self, ev):
        custom_context_menu_event(ev, self)


class Completer(QCompleter):
    """
    From https://gitter.im/baudren/NoteOrganiser?at=55afbefdcce129d570a3c188
    """

    def __init__(self, parent=None):
        super(Completer, self).__init__(parent)

        self.setCaseSensitivity(Qt.CaseInsensitive)
        self.setCompletionMode(QCompleter.PopupCompletion)
        self.setWrapAround(False)

    # Add texts instead of replace
    def pathFromIndex(self, index):
        path = QCompleter.pathFromIndex(self, index)

        lst = str(self.widget().text()).split('$')

        if len(lst) > 1:
            path = '%s%s' % ('$'.join(lst[:-1]), path)

        return path

    # Add operator to separate between texts
    def splitPath(self, path):
        for ch in [' ', '+', '-', '*', '/', '(']:
            path = str(path.split(ch)[-1]).lstrip(' ')
        return [path]



from PyQt5.QtCore import Qt

from pymead.gui.dialogs import PymeadMessageBox


def disp_message_box(message: str, parent, theme: dict, message_mode: str = 'error', rich_text: bool = False):
    window_title_keys = {"error": "Error", "info": "Information", "warn": "Warning"}
    msg_box = PymeadMessageBox(parent, msg=message, window_title=window_title_keys[message_mode], msg_mode=message_mode,
                               theme=theme)
    if rich_text:
        msg_box.setTextFormat(Qt.RichText)
    msg_box.setText(message)
    msg_box.exec()



import numpy as np

from pymead.core.parametric_curve import ParametricCurve, PCurveData
from pymead.core.point import PointSequence, Point
from pymead.utils.nchoosek import nchoosek


class Bezier(ParametricCurve):

    def __init__(self, point_sequence: PointSequence, name: str or None = None, t_start: float = None,
                 t_end: float = None, **kwargs):
        r"""
        Computes the Bézier curve through the control points ``P`` according to

        .. math::

            \vec{C}(t)=\sum_{i=0}^n \vec{P}_i B_{i,n}(t)

        where :math:`B_{i,n}(t)` is the Bernstein polynomial, given by

        .. math::

            B_{i,n}(t)={n \choose i} t^i (1-t)^{n-i}

        Also included are first derivative, second derivative, and curvature information. These are given by

        .. math::

            \vec{C}'(t)=n \sum_{i=0}^{n-1} (\vec{P}_{i+1} - \vec{P}_i B_{i,n-1}(t)

        .. math::

            \vec{C}''(t)=n(n-1) \sum_{i=0}^{n-2} (\vec{P}_{i+2}-2\vec{P}_{i+1}+\vec{P}_i) B_{i,n-2}(t)

        .. math::

            \kappa(t)=\frac{C'_x(t) C''_y(t) - C'_y(t) C''_x(t)}{[(C'_x)^2(t) + (C'_y)^2(t)]^{3/2}}

        Here, the :math:`'` and :math:`''` superscripts are the first and second derivatives with respect to
        :math:`x` and :math:`y`, not the parameter :math:`t`. The result of :math:`\vec{C}''(t)`, for example,
        is a vector with two components, :math:`C''_x(t)` and :math:`C''_y(t)`.

        .. _cubic-bezier:
        .. figure:: ../images/cubic_bezier_light.*
            :class: only-light
            :width: 600
            :align: center

            Cubic Bézier curve

        .. figure:: ../images/cubic_bezier_dark.*
            :class: only-dark
            :width: 600
            :align: center

            Cubic Bézier curve

        An example cubic Bézier curve (degree :math:`n=3`) is shown above. Note that the curve passes
        through the first and last control points and has a local slope at :math:`P_0` equal to the slope of the
        line passing through :math:`P_0` and :math:`P_1`. Similarly, the local slope at :math:`P_3` is equal to
        the slope of the line passing through :math:`P_2` and :math:`P_3`. These properties of Bézier curves allow us to
        easily enforce :math:`G^0` and :math:`G^1` continuity at Bézier curve "joints" (common endpoints of
        connected Bézier curves).

        Parameters
        ----------
        point_sequence: PointSequence
            Sequence of points defining the control points for the Bézier curve

        name: str or ``None``
            Optional name for the curve. Default: ``None``

        t_start: float or ``None``
            Optional starting parameter vector value for the Bézier curve. Not specifying this value automatically
            gives a value of ``0.0``. Default: ``None``

        t_end: float or ``None``
            Optional ending parameter vector value for the Bézier curve. Not specifying this value automatically
            gives a value of ``1.0``. Default: ``None``
        """
        super().__init__(sub_container="bezier", **kwargs)
        self._point_sequence = None
        self.degree = None
        self.set_point_sequence(point_sequence)
        self.t_start = t_start
        self.t_end = t_end
        name = "Bezier-1" if name is None else name
        self.set_name(name)
        self.curve_connections = []
        self._add_references()

    def _add_references(self):
        for idx, point in enumerate(self.point_sequence().points()):
            # If any curves are found at the start point, add their pointers as curve connections
            if idx == 0:
                for curve in point.curves:
                    if not curve.reference:  # Do not include reference curves
                        self.curve_connections.append(curve)

            # If any other curves are found at the end point, add their pointers as curve connections
            elif idx == len(self.point_sequence()) - 1:
                for curve in point.curves:
                    if not curve.reference:  # Do not include reference curves
                        self.curve_connections.append(curve)

            # Add the object reference to each point in the curve
            if self not in point.curves:
                point.curves.append(self)

    def point_sequence(self):
        return self._point_sequence

    def set_point_sequence(self, point_sequence: PointSequence):
        self._point_sequence = point_sequence
        self.degree = len(point_sequence) - 1

    def reverse_point_sequence(self):
        self.point_sequence().reverse()

    def insert_point(self, idx: int, point: Point):
        self.point_sequence().insert_point(idx, point)
        self.degree += 1

    def insert_point_after_point(self, point_to_add: Point, preceding_point: Point):
        idx = self.point_sequence().point_idx_from_ref(preceding_point) + 1
        self.insert_point(idx, point_to_add)
        if self not in point_to_add.curves:
            point_to_add.curves.append(self)
        if self.canvas_item is not None:
            self.canvas_item.point_items.insert(idx, point_to_add.canvas_item)
            self.canvas_item.updateCurveItem(self.evaluate())

    def point_removal_deletes_curve(self):
        return len(self.point_sequence()) <= 3

    def remove_point(self, idx: int or None = None, point: Point or None = None):
        if isinstance(point, Point):
            idx = self.point_sequence().point_idx_from_ref(point)
        self.point_sequence().remove_point(idx)
        self.degree -= 1

        if len(self.point_sequence()) > 2:
            delete_curve = False
        else:
            delete_curve = True

        return delete_curve

    def remove(self):
        if self.canvas_item is not None:
            self.canvas_item.sigRemove.emit(self.canvas_item)

    @staticmethod
    def bernstein_poly(n: int, i: int, t: int or float or np.ndarray):
        r"""
        Calculates the Bernstein polynomial for a given Bézier curve order, index, and parameter vector. The
        Bernstein polynomial is described by

        .. math::

            B_{i,n}(t)={n \choose i} t^i (1-t)^{n-i}

        Arguments
        =========
        n: int
            Bézier curve degree (one less than the number of control points in the Bézier curve)
        i: int
            Bézier curve index
        t: int, float, or np.ndarray
            Parameter vector for the Bézier curve

        Returns
        =======
        np.ndarray
            Array of values of the Bernstein polynomial evaluated for each point in the parameter vector
        """
        return nchoosek(n, i) * t ** i * (1.0 - t) ** (n - i)

    @staticmethod
    def finite_diff_P(P: np.ndarray, k: int, i: int):
        """Calculates the finite difference of the control points as shown in
        https://pages.mtu.edu/~shene/COURSES/cs3621/NOTES/spline/Bezier/bezier-der.html

        Arguments
        =========
        P: np.ndarray
            Array of control points for the Bézier curve
        k: int
            Finite difference level (e.g., k = 1 is the first derivative finite difference)
        i: int
            An index referencing a location in the control point array
        """

        def finite_diff_recursive(_k, _i):
            if _k > 1:
                return finite_diff_recursive(_k - 1, _i + 1) - finite_diff_recursive(_k - 1, _i)
            else:
                return P[_i + 1, :] - P[_i, :]

        return finite_diff_recursive(k, i)

    def derivative(self, P: np.ndarray, t: np.ndarray, degree: int, order: int):
        r"""
        Calculates an arbitrary-order derivative of the Bézier curve

        Parameters
        ==========
        P: np.ndarray
            The control point array

        t: np.ndarray
            The parameter vector

        degree: int
            The degree of the Bézier curve

        order: int
            The derivative order. For example, ``order=2`` returns the second derivative.

        Returns
        =======
        np.ndarray
            An array of ``shape=(N,2)`` where ``N`` is the number of evaluated points specified by the :math:`t` vector.
            The columns represent :math:`C^{(m)}_x(t)` and :math:`C^{(m)}_y(t)`, where :math:`m` is the
            derivative order.
        """
        return np.sum(np.array([np.prod(np.array([degree - idx for idx in range(order)])) *
                                np.array([self.finite_diff_P(P, order, i)]).T *
                                np.array([self.bernstein_poly(degree - order, i, t)])
                                for i in range(degree + 1 - order)]), axis=0).T

    def evaluate(self, t: np.array or None = None, **kwargs):
        r"""
        Evaluates the curve using an optionally specified parameter vector.

        Parameters
        ==========
        t: np.ndarray or ``None``
            Optional direct specification of the parameter vector for the curve. Not specifying this value
            gives a linearly spaced parameter vector from ``t_start`` or ``t_end`` with the default size.
            Default: ``None``

        Returns
        =======
        PCurveData
            Data class specifying the following information about the Bézier curve:

            .. math::

                    C_x(t), C_y(t), C'_x(t), C'_y(t), C''_x(t), C''_y(t), \kappa(t)

            where the :math:`x` and :math:`y` subscripts represent the :math:`x` and :math:`y` components of the
            vector-valued functions :math:`\vec{C}(t)`, :math:`\vec{C}'(t)`, and :math:`\vec{C}''(t)`.
        """
        # Generate the parameter vector
        t = ParametricCurve.generate_t_vec(**kwargs) if t is None else t

        # Number of control points, curve degree, control point array
        n_ctrl_points = len(self.point_sequence())
        degree = n_ctrl_points - 1
        P = self.point_sequence().as_array()

        # Evaluate the curve
        x, y = np.zeros(t.shape), np.zeros(t.shape)
        for i in range(n_ctrl_points):
            # Calculate the x- and y-coordinates of the Bézier curve given the input vector t
            x += P[i, 0] * self.bernstein_poly(degree, i, t)
            y += P[i, 1] * self.bernstein_poly(degree, i, t)
        xy = np.column_stack((x, y))

        # Calculate the first derivative
        first_deriv = self.derivative(P=P, t=t, degree=degree, order=1)
        xp = first_deriv[:, 0]
        yp = first_deriv[:, 1]

        # Calculate the second derivative
        second_deriv = self.derivative(P=P, t=t, degree=degree, order=2)
        xpp = second_deriv[:, 0]
        ypp = second_deriv[:, 1]

        # Combine the derivative x and y data
        xpyp = np.column_stack((xp, yp))
        xppypp = np.column_stack((xpp, ypp))

        # Calculate the curvature
        with np.errstate(divide='ignore', invalid='ignore'):
            # Calculate the curvature of the Bézier curve (k = kappa = 1 / R, where R is the radius of curvature)
            k = np.true_divide((xp * ypp - yp * xpp), (xp ** 2 + yp ** 2) ** (3 / 2))

        # Calculate the radius of curvature: R = 1 / kappa
        with np.errstate(divide='ignore', invalid='ignore'):
            R = np.true_divide(1, k)

        return PCurveData(t=t, xy=xy, xpyp=xpyp, xppypp=xppypp, k=k, R=R)

    def split(self, t_split: float):

        # Number of control points, curve degree, control point array
        n_ctrl_points = len(self.point_sequence())
        degree = n_ctrl_points - 1
        P = self.point_sequence().as_array()

        def de_casteljau(i: int, j: int) -> np.ndarray:
            """
            Based on https://en.wikipedia.org/wiki/De_Casteljau%27s_algorithm. Recursive algorithm where the
            base case is just the value of the ith original control point.

            Parameters
            ----------
            i: int
                Lower index
            j: int
                Upper index

            Returns
            -------
            np.ndarray
                A one-dimensional array containing the :math:`x` and :math:`y` values of a control point evaluated
                at :math:`(i,j)` for a Bézier curve split at the parameter value ``t_split``
            """
            if j == 0:
                return P[i, :]
            return de_casteljau(i, j - 1) * (1 - t_split) + de_casteljau(i + 1, j - 1) * t_split

        bez_split_1_P = np.array([de_casteljau(i=0, j=i) for i in range(n_ctrl_points)])
        bez_split_2_P = np.array([de_casteljau(i=i, j=degree - i) for i in range(n_ctrl_points)])

        if self.geo_col is None:
            bez_1_points = [self.point_sequence().points()[0]] + [Point(*xy.tolist()) for xy in bez_split_1_P[1:, :]]
            bez_2_points = [bez_1_points[-1]] + [Point(*xy.tolist()) for xy in bez_split_2_P[1:-1, :]] + [
                self.point_sequence().points()[-1]]
        else:
            bez_1_points = [self.point_sequence().points()[0]] + [
                self.geo_col.add_point(*xy.tolist()) for xy in bez_split_1_P[1:, :]]
            bez_2_points = [bez_1_points[-1]] + [
                self.geo_col.add_point(*xy.tolist()) for xy in bez_split_2_P[1:-1, :]] + [
                self.point_sequence().points()[-1]]

        bez_1_point_seq = PointSequence(bez_1_points)
        bez_2_point_seq = PointSequence(bez_2_points)

        if self.geo_col is None:
            return (
                Bezier(point_sequence=bez_1_point_seq),
                Bezier(point_sequence=bez_2_point_seq)
            )
        else:
            for point in self.point_sequence().points()[1:-1]:
                self.geo_col.remove_pymead_obj(point)
            return (
                self.geo_col.add_bezier(point_sequence=bez_1_point_seq, name="BezSplit"),
                self.geo_col.add_bezier(point_sequence=bez_2_point_seq, name="BezSplit")
            )

    def get_dict_rep(self):
        return {"points": [pt.name() for pt in self.point_sequence().points()]}


def main():
    points = np.array([
        [0.0, 0.0],
        [0.1, -0.15],
        [0.2, 0.13],
        [0.4, 0.2],
        [0.7, -0.1],
        [0.9, -0.05],
        [1.0, 0.0]
    ])
    original_point_seq = PointSequence.generate_from_array(points)
    bez = Bezier(original_point_seq)
    bez1, bez2 = bez.split(0.7)
    bez_xy = bez.evaluate().xy
    bez1_xy = bez1.evaluate().xy
    bez2_xy = bez2.evaluate().xy

    import matplotlib.pyplot as plt
    plt.plot(bez_xy[:, 0], bez_xy[:, 1], color="steelblue", ls="solid", lw=4)
    plt.plot(bez1_xy[:, 0], bez1_xy[:, 1], color="white", ls="dotted", lw=2)
    plt.plot(bez2_xy[:, 0], bez2_xy[:, 1], color="indianred", ls="dashdot", lw=2)
    plt.plot(original_point_seq.as_array()[:, 0], original_point_seq.as_array()[:, 1], color="grey", ls="dotted", marker="x")
    plt.plot(bez1.point_sequence().as_array()[:, 0], bez1.point_sequence().as_array()[:, 1], color="black", ls="dotted", marker="s")
    plt.plot(bez2.point_sequence().as_array()[:, 0], bez2.point_sequence().as_array()[:, 1], color="gold", ls="dotted", marker="d")
    plt.show()


if __name__ == "__main__":
    main()



import typing
from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
from shapely.geometry import Polygon, LineString

from pymead.core.parametric_curve import INTERMEDIATE_NT
from pymead.core.point import Point
from pymead.core.pymead_obj import PymeadObj
from pymead.core.transformation import Transformation2D
from pymead.post.fonts_and_colors import font
from pymead.post.plot_formatters import format_axis_scientific


class Airfoil(PymeadObj):
    """
    This is a primary class in `pymead`, which defines an airfoil by a leading edge, a trailing edge,
    and optionally an upper-surface endpoint and a lower-surface endpoint in the case of a blunt airfoil. For the
    purposes of single-airfoil evaluation method (such as XFOIL or the built-in panel code), instances of this class
    are sufficient. For multi-element airfoil evaluation (such as MSES), instances of this class are stored in the
    container class, ``pymead.core.mea.MEA``, which adds some additional and necessary functionality. Coordinates
    are stored in the ``coords`` attribute and can be updated using the ``update_coords`` method.
    """
    def __init__(self, leading_edge: Point, trailing_edge: Point,
                 upper_surf_end: Point = None, lower_surf_end: Point = None, name: str or None = None):
        r"""

        Parameters
        ----------
        leading_edge: Point
            The airfoil's leading edge point (usually at :math:`(0,0)` for a typical single airfoil configuration)

        trailing_edge: Point
            The airfoil's trailing edge point (usually at :math:`(1,0)` for a typical single airfoil configuration)

        upper_surf_end: Point
            Optional specification of the upper surface endpoint (the first point in the Selig file format).
            If this point is not specified, the trailing edge point is used instead. Default: ``None``

        lower_surf_end: Point
            Optional specification of the lower surface endpoint (the last point in the Selig file format).
            If this point is not specified, the trailing edge point is used instead. Default: ``None``

        name: str
            Optional name for the airfoil. If ``None``, a default name is used. Default: ``None``
        """

        super().__init__(sub_container="airfoils")

        # Point inputs
        self.leading_edge = leading_edge
        self.trailing_edge = trailing_edge
        self.upper_surf_end = upper_surf_end if upper_surf_end is not None else trailing_edge
        self.lower_surf_end = lower_surf_end if lower_surf_end is not None else trailing_edge

        # Name the airfoil
        name = "Airfoil-1" if name is None else name
        self.set_name(name)

        # Properties to set during closure check
        self.upper_te_curve = None
        self.lower_te_curve = None
        self.curves = []
        self.curves_to_reverse = []

        # Check if the curves in the curve list form a single closed loop
        self.check_closed()

        # Add the airfoil reference to the curves
        for curve in self.curves:
            curve.airfoil = self

        self.coords = self.get_coords_selig_format()

    def check_closed(self):
        """
        This method checks if the airfoil is composed of a closed set of curves.
        If the airfoil is closed, the method passes, but if the airfoil is not closed, a ``ClosureError`` is raised.
        This method also determines which curves need to be evaluated in the opposite parametric direction of their
        point sequences.

        Returns
        -------

        """
        # Get the trailing edge upper curve
        if self.trailing_edge is self.upper_surf_end:
            self.upper_te_curve = None
        else:
            for curve in self.trailing_edge.curves:
                if self.upper_surf_end in curve.point_sequence().points():
                    self.upper_te_curve = curve
                    break
            if self.upper_te_curve is None:
                raise ClosureError("Could not identify the upper trailing edge line/curve")

        # Get the trailing edge lower curve
        if self.trailing_edge is self.lower_surf_end:
            self.lower_te_curve = None
        else:
            for curve in self.trailing_edge.curves:
                if self.lower_surf_end in curve.point_sequence().points():
                    self.lower_te_curve = curve
                    break
            if self.lower_te_curve is None:
                raise ClosureError("Could not identify the lower trailing edge line/curve")

        # Loop through the rest of the curves
        current_curve = None
        if self.upper_te_curve is None:
            current_curve = self.upper_surf_end.curves[0]
        else:
            for curve in self.upper_surf_end.curves:
                if curve is not self.upper_te_curve:
                    current_curve = curve
                    break

        if current_curve is None:
            raise ClosureError("Curve loop is not closed")

        for point in current_curve.point_sequence().points():
            if len(point.curves) > 2:
                raise BranchError("Found more than two curves associated with a curve endpoint in the loop")

        previous_point = self.upper_surf_end
        closed = False
        while not closed:
            self.curves.append(current_curve)

            idx_of_prev_point = current_curve.point_sequence().points().index(previous_point)
            if idx_of_prev_point == 0:
                idx_of_next_point = -1
            else:
                self.curves_to_reverse.append(current_curve)
                idx_of_next_point = 0

            next_point = current_curve.point_sequence().points()[idx_of_next_point]

            if next_point is self.lower_surf_end:
                closed = True
                break

            for curve in next_point.curves:
                if curve is not current_curve:
                    current_curve = curve
                    break

            previous_point = next_point
            pass

        if not closed:
            raise ClosureError("Curve loop not closed")

    def get_coords_selig_format(self, max_airfoil_points: int = None, curvature_exp: float = 2.0) -> np.ndarray:
        r"""
        Gets the coordinates of the airfoil in the Selig file format (coordinate array of size :math:`N \times 2`,
        where :math:`N` is the number of airfoil coordinates, and the columns represent :math:`x` and :math:`y`). The
        order of the points is counter-clockwise, with the start and end at the upper surface trailing edge point and
        lower surface trailing edge point, respectively.

        Parameters
        ----------
        max_airfoil_points: int
            Optional value specifying the maximum number of airfoil points. If this value is left as ``None``,
            no downsampling will be performed. Default: ``None``

        curvature_exp: float
            Optional value specifying the curvature exponent used in the ``downsample`` method. If
            ``max_airfoil_points`` is left as ``None``, this value will be ignored. Default: 2

        Returns
        -------
        np.ndarray
            Coordinate array (size :math:`N \times 2`)

        """
        if max_airfoil_points is not None:
            t_vec_list = self.downsample(max_airfoil_points=max_airfoil_points, curvature_exp=curvature_exp)
        else:
            t_vec_list = [None for _ in self.curves]

        coords = None
        for curve, t_vec in zip(self.curves, t_vec_list):
            p_curve_data = curve.evaluate(t_vec)
            arr = p_curve_data.xy
            if curve in self.curves_to_reverse:
                arr = np.flipud(arr)
            if coords is None:
                coords = arr
            else:
                coords = np.row_stack((coords, arr[1:, :]))
        return coords

    def update_coords(self):
        """
        Updates the coordinates of the airfoil, and passes this data to the canvas item if it exists.

        Returns
        -------

        """
        self.coords = self.get_coords_selig_format()
        if self.canvas_item is not None:
            self.canvas_item.data = self.coords

    def save_coords_selig_format(self, file_name: str) -> None:
        """
        Saves this airfoil's coords in Selig file format: :math:`x`- and :math:`y`-columns in counter-clockwise order,
        starting at the upper surface trailing edge and ending at the lower surface trailing edge. File saves as a
        text file (usually a :code:`.txt` or :code:`.dat` file extension).

        Parameters
        ==========
        file_name: str
            Full path to the coordinate file save location
        """
        coords = self.get_coords_selig_format()
        np.savetxt(file_name, coords)

    def measure_chord(self):
        """
        Measures the chord length

        Returns
        -------
        float
            The airfoil's current chord length

        """
        return self.leading_edge.measure_distance(self.trailing_edge)

    def measure_alpha(self):
        """
        Measures the angle of attack in radians

        Returns
        -------
        float
            The airfoil's current angle of attack

        """
        return -self.leading_edge.measure_angle(self.trailing_edge)

    def get_chord_relative_coords(self, coords: np.ndarray = None, max_airfoil_points: int = None,
                                  curvature_exp: float = 2.0) -> np.ndarray:
        """
        Gets the chord-relative values of the airfoil coordinates. The airfoil is transformed such that the leading
        edge is at :math:`(0,0)` and the trailing edge is at :math:`(1,0)`.

        Parameters
        ----------
        coords: np.ndarray or None
            Optional Selig format airfoil coordinates (only specified if computational speed is important).
            If the coordinates are not specified, they are calculated.

        max_airfoil_points: int
            Optional value specifying the maximum number of airfoil points. If this value is left as ``None``,
            no downsampling will be performed. Default: ``None``

        curvature_exp: float
            Optional value specifying the curvature exponent used in the ``downsample`` method. If
            ``max_airfoil_points`` is left as ``None``, this value will be ignored. Default: 2

        Returns
        -------
        np.ndarray
            An :math:`N \times 2` array of transformed airfoil coordinates
        """
        # Get the chord length and angle of attack
        chord_length = self.measure_chord()
        angle_of_attack = self.measure_alpha()

        # Get the transformation object
        transformation = Transformation2D(
            tx=[-self.leading_edge.x().value()],
            ty=[-self.leading_edge.y().value()],
            r=[angle_of_attack],
            sx=[1 / chord_length],
            sy=[1 / chord_length],
            rotation_units="rad",
            order="t,s,r"
        )

        coords = self.get_coords_selig_format(max_airfoil_points, curvature_exp) if coords is None else coords
        return transformation.transform(coords)

    def get_scaled_coords(self, coords: np.ndarray = None, max_airfoil_points: int = None,
                                curvature_exp: float = 2.0) -> np.ndarray:
        r"""
        Gets the chord-relative values of the airfoil coordinates. The airfoil is transformed such that the leading
        edge is at :math:`(0,0)` and the trailing edge is at :math:`(1,0)`.

        Parameters
        ----------
        coords: np.ndarray or None
            Optional Selig format airfoil coordinates (only specified if computational speed is important).
            If the coordinates are not specified, they are calculated.

        max_airfoil_points: int
            Optional value specifying the maximum number of airfoil points. If this value is left as ``None``,
            no downsampling will be performed. Default: ``None``

        curvature_exp: float
            Optional value specifying the curvature exponent used in the ``downsample`` method. If
            ``max_airfoil_points`` is left as ``None``, this value will be ignored. Default: 2

        Returns
        -------
        np.ndarray
            An :math:`N \times 2` array of transformed airfoil coordinates
        """
        # Get the chord length and angle of attack
        chord_length = self.measure_chord()

        # Get the transformation object
        transformation = Transformation2D(
            tx=[0.0],
            ty=[0.0],
            r=[0.0],
            sx=[1 / chord_length],
            sy=[1 / chord_length],
            rotation_units="rad",
            order="t,s,r"
        )

        coords = self.get_coords_selig_format(max_airfoil_points, curvature_exp) if coords is None else coords
        return transformation.transform(coords)

    @staticmethod
    def convert_coords_to_shapely_format(coords: np.ndarray) -> typing.List[tuple]:
        r"""
        Converts a set of airfoil coordinates to the ``shapely`` native data format (list of tuples).

        Parameters
        ----------
        coords: np.ndarray
            Airfoil coordinates in the form of an :math:`N \times 2` array, where each row represents an :math:`(x,y)`
            coordinate pair

        Returns
        -------
        typing.List[tuple]
            The airfoil coordinates in the ``shapely`` format (a list of tuples, where each tuple has two elements
            representing :math:`x` and :math:`y`)

        """
        return list(map(tuple, coords))

    @staticmethod
    def create_line_string(coords_shapely_format: typing.List[tuple]):
        """
        Creates an instance of the ``shapely.geometry.LineString`` class from an ordered sequence of points in the
        ``shapely`` format.

        Parameters
        ----------
        coords_shapely_format: typing.List[tuple]
            Airfoil coordinates in the ``shapely`` format (list of tuples)

        Returns
        -------
        LineString
            A ``shapely.geometry.LineString`` used to represent the airfoil surface or any other sequence of points

        """
        return LineString(coords_shapely_format)

    @staticmethod
    def create_shapely_polygon(line_string: LineString) -> Polygon:
        r"""
        Creates an instance of the ``shapely.geometry.Polygon`` class from a line string.

        Parameters
        ----------
        line_string: LineString
            A ``shapely.geometry.LineString`` representing a sequence of lines

        Returns
        -------
        Polygon
            A ``shapely.geometry.Polygon`` used to represent the closed shape formed by the airfoil

        """
        return Polygon(line_string)

    def compute_area(self, airfoil_polygon: Polygon = None) -> float:
        """Computes the area of the airfoil as the area of a many-sided polygon enclosed by the airfoil coordinates
        using the `shapely <https://shapely.readthedocs.io/en/stable/manual.html>`_ library.

        Parameters
        ----------
        airfoil_polygon: Polygon or None
            Optional specification of the airfoil's ``shapely.geometry.Polygon`` for speed. If ``None``, it
            will be computed.

        Returns
        -------
        float
            The area of the airfoil
        """
        airfoil_polygon = self.create_shapely_polygon(
            self.create_line_string(
                self.convert_coords_to_shapely_format(
                    self.get_chord_relative_coords(
                        self.get_coords_selig_format()
                    )
                )
            )
        ) if airfoil_polygon is None else airfoil_polygon
        return airfoil_polygon.area

    def check_self_intersection(self, airfoil_line_string: LineString = None) -> bool:
        """Determines whether the airfoil intersects itself using the `is_simple()` function of the
        `shapely <https://shapely.readthedocs.io/en/stable/manual.html>`_ library.

        Parameters
        ----------
        airfoil_line_string: LineString or None
            Optional specification of the airfoil's ``shapely.geometry.LineString`` for speed. If ``None``, it
            will be computed.

        Returns
        -------
        bool
            Describes whether the airfoil intersects itself
        """
        airfoil_line_string = self.create_line_string(
            self.convert_coords_to_shapely_format(
                self.get_chord_relative_coords(
                    self.get_coords_selig_format()
                )
            )
        ) if airfoil_line_string is None else airfoil_line_string
        return not airfoil_line_string.is_simple

    def compute_min_radius(self) -> float:
        """
        Computes the minimum radius of curvature for the airfoil.

        Returns
        -------
        float:
            The minimum radius of curvature

        """
        return min([np.abs(curve.evaluate().R).min() for curve in self.curves])

    def compute_thickness(self, airfoil_line_string: LineString = None, n_lines: int = 201) -> typing.Dict[str, float]:
        r"""Calculates the thickness distribution and maximum thickness of the airfoil.

        Parameters
        ----------
        airfoil_line_string: LineString or None
            Optional specification of the airfoil's ``shapely.geometry.LineString`` for speed. If ``None``, it
            will be computed.

        n_lines: int
          Describes the number of lines evenly spaced along the chordline produced to determine the thickness
          distribution. Default: 201

        Returns
        =======
        dict
          The list of :math:`x`-values used for the thickness distribution calculation, the thickness distribution, the
          maximum value of the thickness distribution, and, if ``return_max_thickness_location=True``,
          the :math:`x/c`-location of the maximum thickness value.
        """
        airfoil_line_string = self.create_line_string(
            self.convert_coords_to_shapely_format(
                self.get_chord_relative_coords(
                    self.get_coords_selig_format()
                )
            )
        ) if airfoil_line_string is None else airfoil_line_string
        x_thickness = np.linspace(0.0, 1.0, n_lines)
        thickness = []
        for idx in range(n_lines):
            line_string = LineString([(x_thickness[idx], -1), (x_thickness[idx], 1)])
            x_inters = line_string.intersection(airfoil_line_string)
            if x_inters.is_empty:
                thickness.append(0.0)
            else:
                thickness.append(x_inters.convex_hull.length)
        x_thickness = x_thickness
        thickness = thickness
        max_thickness = max(thickness)
        x_c_loc_idx = np.argmax(thickness)
        x_c_loc = x_thickness[x_c_loc_idx]
        return {
            "x/c": x_thickness,
            "t/c": thickness,
            "t/c_max": max_thickness,
            "t/c_max_x/c_loc": x_c_loc
        }

    def compute_thickness_at_points(self, x_over_c: np.ndarray, airfoil_line_string: LineString = None,
                                    start_y_over_c: float = -1.0, end_y_over_c: float = 1.0) -> np.ndarray:
        """
        Calculates the thickness (t/c) at a set of x-locations (x/c)

        Parameters
        ----------
        x_over_c: float or list or np.ndarray
            The :math:`x/c` locations at which to evaluate the thickness

        airfoil_line_string: LineString or None
            Optional specification of the airfoil's ``shapely.geometry.LineString`` for speed. If ``None``, it
            will be computed.

        start_y_over_c: float
            The :math:`y/c` location to draw the first point in a line whose intersection with the airfoil is checked.
            May need to decrease this value for unusually thick airfoils

        end_y_over_c: float
            The :math:`y/c` location to draw the last point in a line whose intersection with the airfoil is checked.
            May need to increase this value for unusually thick airfoils

        Returns
        -------
        np.ndarray
            An array of thickness (:math:`t/c`) values corresponding to the input :math:`x/c` values
        """
        airfoil_line_string = self.create_line_string(
            self.convert_coords_to_shapely_format(
                self.get_chord_relative_coords(
                    self.get_coords_selig_format()
                )
            )
        ) if airfoil_line_string is None else airfoil_line_string
        thickness = np.array([])
        for pt in x_over_c:
            line_string = LineString([(pt, start_y_over_c), (pt, end_y_over_c)])
            x_inters = line_string.intersection(airfoil_line_string)
            if pt == 1.0:
                thickness = np.append(thickness, self.lower_surf_end.measure_distance(self.upper_surf_end))
            elif x_inters.is_empty:  # If no intersection between line and airfoil LineString,
                thickness = np.append(thickness, 0.0)
            else:
                thickness = np.append(thickness, x_inters.convex_hull.length)
        return thickness  # Return an array of t/c values corresponding to the x/c locations

    def compute_camber_at_points(self, x_over_c: np.ndarray, airfoil_line_string: LineString = None,
                                 start_y_over_c: float = -1.0, end_y_over_c: float = 1.0) -> np.ndarray:
        """Calculates the thickness (t/c) at a set of x-locations (x/c)

        Parameters
        ----------
        x_over_c: float or list or np.ndarray
            The :math:`x/c` locations at which to evaluate the camber

        airfoil_line_string: LineString or None
            Optional specification of the airfoil's ``shapely.geometry.LineString`` for speed. If ``None``, it
            will be computed.

        start_y_over_c: float
            The :math:`y/c` location to draw the first point in a line whose intersection with the airfoil is checked.
            May need to decrease this value for unusually thick airfoils

        end_y_over_c: float
            The :math:`y/c` location to draw the last point in a line whose intersection with the airfoil is checked.
            May need to increase this value for unusually thick airfoils

        Returns
        -------
        np.ndarray
            An array of thickness (:math:`t/c`) values corresponding to the input :math:`x/c` values
        """
        airfoil_line_string = self.create_line_string(
            self.convert_coords_to_shapely_format(
                self.get_chord_relative_coords(
                    self.get_coords_selig_format()
                )
            )
        ) if airfoil_line_string is None else airfoil_line_string
        camber = np.array([])
        for pt in x_over_c:
            line_string = LineString([(pt, start_y_over_c), (pt, end_y_over_c)])
            x_inters = line_string.intersection(airfoil_line_string)
            if pt == 0.0 or pt == 1.0 or x_inters.is_empty:
                camber = np.append(camber, 0.0)
            else:
                camber = np.append(camber, x_inters.convex_hull.centroid.xy[1])
        return camber  # Return an array of h/c values corresponding to the x/c locations

    def contains_point(self, point: np.ndarray, airfoil_polygon: Polygon = None) -> bool:
        """Determines whether a point is contained inside the airfoil

        Parameters
        ----------
        point: np.ndarray or list
            The point to test. Should be either a 1-D ``ndarray`` of the format ``array([<x_val>,<y_val>])`` or
            a list of the format ``[<x_val>,<y_val>]``

        airfoil_polygon: Polygon or None
            Optional specification of the airfoil's ``shapely.geometry.Polygon`` for speed. If ``None``, it
            will be computed.

        Returns
        -------
        bool
            Whether the point is contained inside the airfoil
        """
        airfoil_polygon = self.create_shapely_polygon(
            self.create_line_string(
                self.convert_coords_to_shapely_format(
                    self.get_chord_relative_coords(
                        self.get_coords_selig_format()
                    )
                )
            )
        ) if airfoil_polygon is None else airfoil_polygon
        return airfoil_polygon.contains(Point(point[0], point[1]))

    def contains_line_string(self, points: np.ndarray or list, airfoil_polygon: Polygon = None) -> bool:
        """
        Whether a connected string of points is contained the airfoil

        Parameters
        ----------
        points: np.ndarray or list
            Should be a 2-D array or list of the form ``[[<x_val_1>, <y_val_1>], [<x_val_2>, <y_val_2>], ...]``

        airfoil_polygon: Polygon or None
            Optional specification of the airfoil's ``shapely.geometry.Polygon`` for speed. If ``None``, it
            will be computed.

        Returns
        -------
        bool
            Whether the line string is contained inside the airfoil
        """
        airfoil_polygon = self.create_shapely_polygon(
            self.create_line_string(
                self.convert_coords_to_shapely_format(
                    self.get_chord_relative_coords(
                        self.get_coords_selig_format()
                    )
                )
            )
        ) if airfoil_polygon is None else airfoil_polygon
        line_string = LineString(list(map(tuple, points)))
        return airfoil_polygon.contains(line_string)

    def downsample(self, max_airfoil_points: int, curvature_exp: float = 2.0):
        r"""
        Downsamples the airfoil coordinates based on a curvature exponent. This method works by evaluating each
        Bézier curve using a set number of points (150) and then calculating
        :math:`\mathbf{R_e} = \mathbf{R}^{1/e_c}`, where :math:`\mathbf{R}` is the radius of curvature vector and
        :math:`e_c` is the curvature exponent (an input to this method). Then, :math:`\mathbf{R_e}` is
        normalized by its maximum value and concatenated to a single array for all curves in a given airfoil.
        Finally, ``max_airfoil_points`` are chosen from this array to create a new set of parameter vectors
        for the airfoil.

        Parameters
        ----------
        max_airfoil_points: int
            Maximum number of points in the airfoil (the actual number in the final airfoil may be slightly less)

        curvature_exp: float
            Curvature exponent used to scale the radius of curvature. Values close to 0 place high emphasis on
            curvature, while values close to :math:`\infty` place low emphasis on curvature (creating nearly
            uniform spacing)

        Returns
        -------
        list[np.ndarray]
            List of parameter vectors (one for each Bézier curve)
        """

        if max_airfoil_points > sum([INTERMEDIATE_NT for _ in self.curves]):
            return [None for _ in self.curves]

        nt = np.ceil(max_airfoil_points / len(self.curves)).astype(int)
        curve_data_list = [curve.evaluate(t=np.linspace(0, 1, nt)) for curve in self.curves]

        new_param_vec_list = []
        new_t_concat = np.array([])

        for c_idx, curve_data in enumerate(curve_data_list):
            temp_R = deepcopy(curve_data.R)
            for r_idx, r in enumerate(temp_R):
                if np.isinf(r) and r > 0:
                    temp_R[r_idx] = 10000
                elif np.isinf(r) and r < 0:
                    temp_R[r_idx] = -10000

            exp_R = np.abs(temp_R) ** (1 / curvature_exp)
            new_t = np.zeros(exp_R.shape)
            for i in range(1, new_t.shape[0]):
                new_t[i] = new_t[i - 1] + (exp_R[i] + exp_R[i - 1]) / 2
            new_t = new_t / np.max(new_t)
            new_t_concat = np.concatenate((new_t_concat, new_t))

        indices_to_select = np.linspace(0, new_t_concat.shape[0] - 1,
                                        max_airfoil_points - 2 * len(self.curves)).astype(int)

        t_old = 0.0
        for selection_idx in indices_to_select:
            t = new_t_concat[selection_idx]

            if t == 0.0 and selection_idx == 0:
                new_param_vec_list.append(np.array([0.0]))
            elif t < t_old:
                if t_old != 1.0:
                    new_param_vec_list[-1] = np.append(new_param_vec_list[-1], 1.0)
                if t == 0.0:
                    new_param_vec_list.append(np.array([]))
                else:
                    new_param_vec_list.append(np.array([0.0]))
                new_param_vec_list[-1] = np.append(new_param_vec_list[-1], t)
                t_old = t
            else:
                new_param_vec_list[-1] = np.append(new_param_vec_list[-1], t)
                t_old = t

        return new_param_vec_list

    def plot(self, show: bool = True, save_file: str or None = None, ax: plt.Axes or None = None):
        """
        Plots the airfoil to a ``matplotlib`` figure.

        Parameters
        ----------
        show: bool
            Whether to immediately show the airfoil plot. Default: ``True``

        save_file: str or None
            Name of the file to save. If ``None``, the airfoil image will not be saved to file. Default: ``None``

        ax: plt.Axes or None
            Matplotlib Axes object on which the airfoil will be plotted. Default: ``None``
        """
        if ax is not None:
            fig = ax.figure
        else:
            fig, ax = plt.subplots(figsize=(10, 2))

        # Plot the curves
        for curve in self.curves:
            curve_data = curve.evaluate()
            ax.plot(curve_data.xy[:, 0], curve_data.xy[:, 1], color="steelblue")

        # Plot settings
        ax.set_aspect("equal")
        ax.set_xlabel("x", fontdict=font)
        ax.set_ylabel("y", fontdict=font)
        format_axis_scientific(ax=ax)

        # Save and/or show
        if save_file is not None:
            fig.savefig(save_file, bbox_inches="tight")
        if show:
            plt.show()

    def get_dict_rep(self):
        return {"leading_edge": self.leading_edge.name(), "trailing_edge": self.trailing_edge.name(),
                "upper_surf_end": self.upper_surf_end.name(), "lower_surf_end": self.lower_surf_end.name()}


class BranchError(Exception):
    pass


class ClosureError(Exception):
    pass



import typing


def get_prefix_and_index_from_string(s: str):
    prefix = ''
    idx_str = ''
    for ch in s:
        if ch.isdigit():
            idx_str += ch  # String representation of the FreePoint or AnchorPoint index
        elif ch.isalpha():
            prefix += ch  # Final result of the addition will be 'FP' for a FreePoint or 'AP' for an AnchorPoint
        else:
            raise ValueError(f"Invalid (non alpha-numeric) character found in string {s}")
    idx = int(idx_str)  # Integer representation of the FreePoint or AnchorPoint index
    return prefix, idx


def decrement_string_index(s: str):
    # Get integer index from string
    prefix, idx = get_prefix_and_index_from_string(s)
    # Return a decremented version of the string index
    return prefix + str(idx - 1)


def increment_string_index(s: str):
    # Get integer index from string
    prefix, idx = get_prefix_and_index_from_string(s)
    # Return an incremented version of the string index
    return prefix + str(idx + 1)


def max_string_index_plus_one(str_list: typing.List[str]):
    if str_list and len(str_list) > 0:
        # print(f"{str_list = }")
        idx_list = [int([ch for ch in s if ch.isnumeric()][0]) for s in str_list]
        return 'FP' + str(max(idx_list) + 1)
    else:
        return 'FP0'



import os
import typing
import sys
from copy import deepcopy
from functools import partial

import numpy as np
import pyqtgraph as pg
from PyQt5 import QtWidgets
from PyQt5.QtCore import pyqtSignal, QEventLoop, Qt
from PyQt5.QtGui import QFont, QBrush, QColor
from PyQt5.QtWidgets import QApplication

from pymead.utils.get_airfoil import AirfoilNotFoundError
from pymead.core.line import PolyLine
from pymead.gui.dialogs import PlotExportDialog, WebAirfoilDialog, SplitPolylineDialog
from pymead.gui.polygon_item import PolygonItem

from pymead.core.airfoil import Airfoil

from pymead.core.constraints import *
from pymead.core.constraint_equations import *
from pymead.core.geometry_collection import GeometryCollection
from pymead.core.param import LengthParam
from pymead.core.parametric_curve import ParametricCurve
from pymead.core.point import PointSequence, Point
from pymead.core.pymead_obj import PymeadObj
from pymead.gui.constraint_items import *
from pymead.gui.hoverable_curve import HoverableCurve
from pymead.gui.draggable_point import DraggablePoint
from pymead.utils.read_write_files import load_data
from pymead.utils.misc import get_setting
from pymead.core import UNITS
from pymead import q_settings, GUI_SETTINGS_DIR
from pymead.resources.cmcrameri.cmaps import BERLIN, VIK

q_settings_descriptions = load_data(os.path.join(GUI_SETTINGS_DIR, "q_settings_descriptions.json"))


class AirfoilCanvas(pg.PlotWidget):
    sigEnterPressed = pyqtSignal()
    sigEscapePressed = pyqtSignal()
    sigStatusBarUpdate = pyqtSignal(str, int)

    def __init__(self, parent, geo_col: GeometryCollection, gui_obj):
        super().__init__(parent)
        self.dialog = None
        self.setMenuEnabled(False)
        self.setAspectLocked(True)
        self.disableAutoRange()
        self.points = []
        self.airfoil_text_items = {}
        self.drawing_object = None
        self.creating_collinear_constraint = None
        self.curve_hovered_item = None
        self.point_hovered_item = None
        self.constraint_hovered_item = None
        self.point_text_item = None
        self.enter_connection = None
        self.color_bar_data = None
        self.geo_col = geo_col
        self.geo_col.canvas = self
        self.gui_obj = gui_obj
        self.plot = self.getPlotItem()
        self.setMinimumWidth(500)
        self.setMinimumHeight(300)

    def showPymeadObjs(self, sub_container: str):
        for pymead_obj in self.geo_col.container()[sub_container].values():
            if pymead_obj.canvas_item is None:
                continue
            pymead_obj.canvas_item.show()

    def hidePymeadObjs(self, sub_container: str):
        for pymead_obj in self.geo_col.container()[sub_container].values():
            if pymead_obj.canvas_item is None:
                continue
            pymead_obj.canvas_item.hide()

    def showAllPymeadObjs(self):
        sub_containers = ("points", "bezier", "lines", "airfoils", "geocon")
        for sub_container in sub_containers:
            self.showPymeadObjs(sub_container)
        return {sub_container: True for sub_container in sub_containers}

    def hideAllPymeadObjs(self):
        sub_containers = ("points", "bezier", "lines", "airfoils", "geocon")
        for sub_container in sub_containers:
            self.hidePymeadObjs(sub_container)
        return {sub_container: False for sub_container in sub_containers}

    def setAxisLabels(self, theme: dict):
        label_font = f"{get_setting('axis-label-point-size')}pt {get_setting('axis-label-font-family')}"
        self.plot.setLabel(axis="bottom", text=f"x [{UNITS.current_length_unit()}]", font=label_font,
                           color=theme["main-color"])
        self.plot.setLabel(axis="left", text=f"y [{UNITS.current_length_unit()}]", font=label_font,
                           color=theme["main-color"])
        tick_font = QFont(get_setting("axis-tick-font-family"), get_setting("axis-tick-point-size"))
        self.plot.getAxis("bottom").setTickFont(tick_font)
        self.plot.getAxis("left").setTickFont(tick_font)
        self.plot.getAxis("bottom").setTextPen(theme["main-color"])
        self.plot.getAxis("left").setTextPen(theme["main-color"])

    def toggleGrid(self):
        x_state = self.plot.ctrl.xGridCheck.checkState()
        y_state = self.plot.ctrl.yGridCheck.checkState()
        if x_state or y_state:
            self.plot.showGrid(x=False, y=False)
        else:
            self.plot.showGrid(x=True, y=True)

    def getPointRange(self):
        """
        Gets the minimum and maximum :math:`x` and :math:`y` values of all the points in the GeometryCollection.

        Returns
        typing.Tuple[list]
            x-range and y-range of the points (two-element tuple of two-element lists)
        """
        point_seq = PointSequence(points=[pt for pt in self.geo_col.container()["points"].values()])
        pseq_arr = point_seq.as_array()
        if len(pseq_arr) > 0:
            min_x = pseq_arr[:, 0].min()
            max_x = pseq_arr[:, 0].max()
            min_y = pseq_arr[:, 1].min()
            max_y = pseq_arr[:, 1].max()
        else:
            min_x = -0.05
            max_x = 1.05
            min_y = -0.25
            max_y = 0.25
        return [min_x, max_x], [min_y, max_y]

    def drawPoint(self, x, y):
        """
        Draws a point on the airfoil canvas at a specified :math:`x`- and :math:`y`-location.

        Parameters
        ==========
        x: list
            One-element list where the element represents the :math:`x`-location

        y: list
            One-element list where the element represents the :math:`y`-location
        """
        self.geo_col.add_point(x[0], y[0])

    def addPymeadCanvasItem(self, pymead_obj: PymeadObj):
        # Type-specific actions
        if isinstance(pymead_obj, Point):
            # Create the canvas item
            point_gui = DraggablePoint()
            point_gui.setData(pos=np.array([[pymead_obj.x().value(), pymead_obj.y().value()]]), adj=None,
                              pen=pg.mkPen(color=q_settings.value("scatter_default_pen_color",
                                                                  q_settings_descriptions["scatter_default_pen_color"][
                                                                      1])),
                              pxMode=True, hoverable=True, tip=None)
            point_gui.setZValue(100)

            # Establish a two-way connection between the point data structure and the GUI representation
            pymead_obj.canvas_item = point_gui
            point_gui.point = pymead_obj

            # Set the style
            point_gui.setScatterStyle("default")

            # Connect signals
            point_gui.sigPointClicked.connect(self.pointClicked)
            point_gui.sigPointHovered.connect(self.pointHovered)
            point_gui.sigPointLeaveHovered.connect(self.pointLeaveHovered)
            point_gui.sigPointStartedMoving.connect(self.pointStartedMoving)
            point_gui.sigPointMoving.connect(self.pointMoving)
            point_gui.sigPointFinishedMoving.connect(self.pointFinishedMoving)

            # Add the point to the plot
            self.addItem(point_gui)

        elif isinstance(pymead_obj, ParametricCurve):
            # Create the canvas item
            curve_item = HoverableCurve(curve_type=pymead_obj.curve_type)

            # Establish a two-way connection between the curve data structure and the GUI representation
            pymead_obj.canvas_item = curve_item
            curve_item.parametric_curve = pymead_obj

            # Set the curve style
            curve_item.setCurveStyle("default")

            # Set the curve to be clickable within a specified radius
            curve_item.setClickable(True, width=4)

            # Connect hover/not hover signals
            curve_item.sigCurveHovered.connect(self.curveHovered)
            curve_item.sigCurveNotHovered.connect(self.curveLeaveHovered)
            curve_item.sigLineItemAdded.connect(self.onLineItemAdded)
            curve_item.sigRemove.connect(self.removeCurve)

            # Update the curve data based on the selected control points
            pymead_obj.update()

            # Add the curve to the plot
            self.addItem(curve_item)

        elif isinstance(pymead_obj, GeoCon):

            constraint_item = getattr(
                sys.modules[__name__], f"{type(pymead_obj).__name__}Item")(
                pymead_obj, self.gui_obj.themes[self.gui_obj.current_theme])
            # raise NotImplementedError(f"Constraint {pymead_obj.__class__.__name__} does not yet have a canvas item")

            constraint_item.addItems(self)

            # Connect hover/not hover signals
            # TODO: need to connect these

        elif isinstance(pymead_obj, Airfoil):

            pymead_obj.canvas_item = PolygonItem(data=pymead_obj.coords, airfoil=pymead_obj, gui_obj=self.gui_obj)
            self.addItem(pymead_obj.canvas_item)

            # Connect signals
            pymead_obj.canvas_item.sigPolyEnter.connect(self.airfoil_hovered)
            pymead_obj.canvas_item.sigPolyExit.connect(self.airfoil_exited)

    def closeEnterCallbackConnection(self):
        if self.enter_connection is None:
            return

        self.sigEnterPressed.disconnect(self.enter_connection)
        self.enter_connection = None

    @staticmethod
    def undoRedoAction(action: typing.Callable):
        def wrapped(self, *args, **kwargs):
            self.gui_obj.undo_stack.append(deepcopy(self.geo_col.get_dict_rep()))
            action(self, *args, **kwargs)

        return wrapped

    @staticmethod
    def runSelectionEventLoop(drawing_object: str, starting_message: str, enter_callback: typing.Callable = None):
        drawing_object = drawing_object
        starting_message = starting_message

        def decorator(action: typing.Callable):
            def wrapped(self, *args, **kwargs):
                self.drawing_object = drawing_object
                self.sigStatusBarUpdate.emit(starting_message, 0)
                self.closeEnterCallbackConnection()
                loop = QEventLoop()
                if enter_callback:
                    self.enter_connection = self.sigEnterPressed.connect(partial(enter_callback, self))
                else:
                    self.sigEnterPressed.connect(loop.quit)
                self.sigEscapePressed.connect(loop.quit)
                loop.exec()
                # if len(self.geo_col.selected_objects["points"]) > 0:

                try:
                    action(self, *args, **kwargs)
                except ValueError as e:
                    self.gui_obj.disp_message_box(str(e))

                self.clearSelectedObjects()
                # elif len(self.geo_col.selected_objects["airfoils"]) > 0:
                #     action(self, *args, **kwargs)
                #     self.clearSelectedObjects()
                self.drawing_object = None
                self.sigStatusBarUpdate.emit("", 0)
                if enter_callback:
                    self.closeEnterCallbackConnection()
            return wrapped
        return decorator

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="Points", starting_message="Left click on the canvas to draw a point. "
                                                                     "Press Escape to stop drawing points.")
    def drawPoints(self):
        pass

    @undoRedoAction
    def drawBezierNoEvent(self):
        if len(self.geo_col.selected_objects["points"]) < 2:
            msg = f"Choose at least 2 points to define a curve"
            self.sigStatusBarUpdate.emit(msg, 2000)
            return

        point_sequence = PointSequence([pt for pt in self.geo_col.selected_objects["points"]])
        self.geo_col.add_bezier(point_sequence=point_sequence)

        self.clearSelectedObjects()
        self.sigStatusBarUpdate.emit("Select the first Bezier control point of the next curve", 0)

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="Bezier", starting_message="Select the first Bezier control point")
    def drawBezier(self):
        if len(self.geo_col.selected_objects["points"]) < 2:
            msg = f"Choose at least 2 points to define a curve"
            self.sigStatusBarUpdate.emit(msg, 2000)
            return

        point_sequence = PointSequence([pt for pt in self.geo_col.selected_objects["points"]])
        self.geo_col.add_bezier(point_sequence=point_sequence)

    @runSelectionEventLoop(drawing_object="Beziers", starting_message="Select the first Bezier control point",
                           enter_callback=drawBezierNoEvent)
    def drawBeziers(self):
        if len(self.geo_col.selected_objects["points"]) < 2:
            msg = f"Choose at least 2 points to define a curve"
            self.sigStatusBarUpdate.emit(msg, 2000)
            return

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="LineSegment", starting_message="Select the first line endpoint")
    def drawLineSegment(self):
        if len(self.geo_col.selected_objects["points"]) < 2:
            msg = f"Choose at least 2 points to define a curve"
            self.sigStatusBarUpdate.emit(msg, 2000)
            return

        point_sequence = PointSequence([pt for pt in self.geo_col.selected_objects["points"]])
        self.geo_col.add_line(point_sequence=point_sequence)

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="LineSegments", starting_message="Select the first line endpoint")
    def drawLines(self):
        if len(self.geo_col.selected_objects["points"]) < 2:
            msg = f"Choose at least 2 points to define a curve"
            self.sigStatusBarUpdate.emit(msg, 2000)
            return

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="Airfoil", starting_message="Select the leading edge point")
    def generateAirfoil(self):
        if len(self.geo_col.selected_objects["points"]) not in [2, 4]:
            self.sigStatusBarUpdate.emit(
                "Choose either 2 points (sharp trailing edge) or 4 points (blunt trailing edge)"
                " to generate an airfoil", 4000)
            return

        le = self.geo_col.selected_objects["points"][0]
        te = self.geo_col.selected_objects["points"][1]
        if len(self.geo_col.selected_objects["points"]) > 2:
            upper_surf_end = self.geo_col.selected_objects["points"][2]
            lower_surf_end = self.geo_col.selected_objects["points"][3]
        else:
            upper_surf_end = te
            lower_surf_end = te

        self.geo_col.add_airfoil(leading_edge=le, trailing_edge=te, upper_surf_end=upper_surf_end,
                                 lower_surf_end=lower_surf_end)
        self.gui_obj.permanent_widget.updateAirfoils()

    @undoRedoAction
    def generateWebAirfoil(self, dialog_test_action: typing.Callable = None):
        """
        Generates an airfoil from Airfoil Tools or from a file.

        Parameters
        ----------
        dialog_test_action: typing.Callable or None
            If not ``None``, should be a function that accepts a ``WebAirfoilDialog`` as its sole argument and returns
            nothing. Default: ``None``
        """
        self.dialog = WebAirfoilDialog(self, theme=self.gui_obj.themes[self.gui_obj.current_theme])
        if (dialog_test_action is not None and not dialog_test_action(self.dialog)) or self.dialog.exec_():
            try:
                polyline = self.geo_col.add_polyline(source=self.dialog.value())
            except AirfoilNotFoundError as e:
                self.gui_obj.disp_message_box(f"{e}", message_mode="error")
                return
            polyline.add_polyline_airfoil()
        self.gui_obj.permanent_widget.updateAirfoils()

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="MEA", starting_message="Select the first airfoil")
    def generateMEA(self):
        if len(self.geo_col.selected_objects["airfoils"]) == 0:
            self.sigStatusBarUpdate.emit("Must choose at least 1 airfoil for a multi-element airfoil (MEA) object",
                                         4000)
            return

        self.geo_col.add_mea(airfoils=self.geo_col.selected_objects["airfoils"].copy())

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="DistanceConstraint", starting_message="Select the first point")
    def addDistanceConstraint(self):
        if len(self.geo_col.selected_objects["points"]) != 2:
            self.sigStatusBarUpdate.emit("Choose exactly two points to define a distance constraint", 4000)
            return

        # p1 = self.geo_col.selected_objects["points"][0]
        # p2 = self.geo_col.selected_objects["points"][1]
        #
        # cnstr = self.geo_col.add_distance_constraint(p1, p2)
        # cnstr.add_constraint_to_gcs()

        constraint = DistanceConstraint(*self.geo_col.selected_objects["points"])
        self.geo_col.add_constraint(constraint)

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="RelAngle3Constraint", starting_message="Select any point other than "
                                                                                  "the vertex")
    def addRelAngle3Constraint(self):
        if len(self.geo_col.selected_objects["points"]) != 3:
            msg = (f"Choose exactly three points (start, vertex, and end) for a "
                   f"relative angle 3 constraint")
            self.sigStatusBarUpdate.emit(msg, 4000)
            return

        constraint = RelAngle3Constraint(*self.geo_col.selected_objects["points"])
        self.geo_col.add_constraint(constraint)

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="AntiParallel3Constraint", starting_message="Select any point other than "
                                                                                      "the vertex")
    def addAntiParallel3Constraint(self):
        if len(self.geo_col.selected_objects["points"]) == 3:
            case = 0
        elif len(self.geo_col.selected_objects["points"]) == 2 and len(self.geo_col.selected_objects["polylines"]) == 1:
            case = 1
        else:
            msg = (f"Choose exactly three points (start, vertex, and end) or a curve and two points for an "
                   f"anti-parallel 3 constraint")
            self.sigStatusBarUpdate.emit(msg, 4000)
            return
        if case == 0:
            constraint = AntiParallel3Constraint(*self.geo_col.selected_objects["points"])
        else:
            data = self.geo_col.selected_objects["polylines"][0].evaluate()
            # if (np.isclose(self.geo_col.selected_objects["points"][0].x().value(), data.xy[0, 0]) and
            #     np.isclose(self.geo_col.selected_objects["points"][0].y().value(), data.xy[0, 1])):
            if self.geo_col.selected_objects["points"][0].is_coincident(Point(*data.xy[0, :])):
                point = self.geo_col.add_point(data.xy[1, 0], data.xy[1, 1])
                point_is_first_arg = True
            elif self.geo_col.selected_objects["points"][0].is_coincident(Point(*data.xy[-1, :])):
                point = self.geo_col.add_point(data.xy[-2, 0], data.xy[-2, 1])
                point_is_first_arg = True
            elif self.geo_col.selected_objects["points"][1].is_coincident(Point(*data.xy[0, :])):
                point = self.geo_col.add_point(data.xy[1, 0], data.xy[1, 1])
                point_is_first_arg = False
            else:
                point = self.geo_col.add_point(data.xy[-2, 0], data.xy[-2, 1])
                point_is_first_arg = False

            args = [point, *self.geo_col.selected_objects["points"]] if point_is_first_arg else \
                [*self.geo_col.selected_objects["points"], point]
            constraint = AntiParallel3Constraint(*args, polyline=self.geo_col.selected_objects["polylines"][0],
                                                 point_on_curve=point)
        self.geo_col.add_constraint(constraint)

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="SymmetryConstraint", starting_message="Select the start point of the "
                                                                                 "mirror axis")
    def addSymmetryConstraint(self):
        if len(self.geo_col.selected_objects["points"]) != 4:
            msg = (f"Choose exactly four points (mirror axis start, mirror axis end, tool point, and target point) "
                   f"for a symmetry constraint")
            self.sigStatusBarUpdate.emit(msg, 4000)
            return

        constraint = SymmetryConstraint(*self.geo_col.selected_objects["points"])
        self.geo_col.add_constraint(constraint)

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="Perp3Constraint", starting_message="Select the first point (not the vertex)")
    def addPerp3Constraint(self):
        if len(self.geo_col.selected_objects["points"]) != 3:
            msg = f"Choose exactly three points (start, vertex, and end) for a Perp3Constraint"
            self.sigStatusBarUpdate.emit(msg, 4000)
            return
        constraint = Perp3Constraint(*self.geo_col.selected_objects["points"])
        self.geo_col.add_constraint(constraint)

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="ROCurvatureConstraint", starting_message="Select the curve joint")
    def addROCurvatureConstraint(self):
        if len(self.geo_col.selected_objects["points"]) != 1:
            msg = (f"Choose exactly one point (the curve joint) for a "
                   f"radius of curvature constraint")
            self.sigStatusBarUpdate.emit(msg, 4000)
            return

        constraint = ROCurvatureConstraint(*self.geo_col.selected_objects["points"])
        self.geo_col.add_constraint(constraint)

    @undoRedoAction
    @runSelectionEventLoop(drawing_object="BezierAddPoint",
                           starting_message="First, click the point to add to the curve")
    def addPointToCurve(self, curve_item: HoverableCurve):
        if len(self.geo_col.selected_objects["points"]) != 2:
            msg = (f"Choose exactly two points (the point to add, then the preceding point in the curve) to add"
                   f" a point to a Bezier curve")
            self.sigStatusBarUpdate.emit(msg, 4000)
            return
        bezier_curve = curve_item.parametric_curve
        point_to_add = self.geo_col.selected_objects["points"][0]
        preceding_point = self.geo_col.selected_objects["points"][1]
        bezier_curve.insert_point_after_point(point_to_add, preceding_point)

    @undoRedoAction
    def splitPoly(self, curve_item: HoverableCurve):
        polyline = curve_item.parametric_curve
        dialog = SplitPolylineDialog(self, theme=self.gui_obj.themes[self.gui_obj.current_theme], polyline=polyline,
                                     geo_col=self.geo_col)
        if dialog.exec_():
            self.geo_col.split_polyline(polyline, dialog.value())

    def appendSelectedPoint(self, plot_data_item: pg.PlotDataItem):
        self.geo_col.selected_objects["points"].append(plot_data_item.point)

    def setColorBarData(self, cmap_dict, color_bar, current_theme: str, flow_var: str, min_level_default: float,
                       max_level_default: float):
        self.color_bar_data = dict(cmap_dict=cmap_dict, color_bar=color_bar, current_theme=current_theme,
                                   flow_var=flow_var, min_level_default=min_level_default,
                                   max_level_default=max_level_default)

    def setColorBarLevels(self, min_level: float = None, max_level: float = None):

        if self.color_bar_data["current_theme"] == "dark":
            color_data = BERLIN
        elif self.color_bar_data["current_theme"] == "light":
            color_data = VIK
        else:
            raise ValueError("Could not find color map for the current theme")

        levels = [self.color_bar_data["min_level_default"], self.color_bar_data["max_level_default"]]
        if min_level is not None:
            levels[0] = min_level
        if max_level is not None:
            levels[1] = max_level

        if levels[0] >= levels[1]:
            raise ValueError("Minimum value cannot be greater than or equal to the maximum value")
        if levels[1] <= levels[0]:
            raise ValueError("Maximum value cannot be less than or equal to the minimum value")

        if self.color_bar_data["flow_var"] in ["Cp", "v"]:
            stop = (0.0 - levels[0]) / (levels[1] - levels[0])
            pos = np.linspace(0.0, stop, color_data.shape[0] // 2 + 1)
            pos = pos[:-1]
            pos2 = np.linspace(stop, 1.0, color_data.shape[0] // 2)
            for p in pos2:
                pos = np.append(pos, p)
        else:
            stop = (1.0 - levels[0]) / (levels[1] - levels[0])
            pos = np.linspace(0.0, stop, color_data.shape[0] // 2 + 1)
            pos = pos[:-1]
            pos2 = np.linspace(stop, 1.0, color_data.shape[0] // 2)
            for p in pos2:
                pos = np.append(pos, p)

        self.color_bar_data["cmap_dict"]["dark"] = pg.ColorMap(name="berlin", pos=pos, color=255 * BERLIN + 0.5)
        self.color_bar_data["cmap_dict"]["light"] = pg.ColorMap(name="vik", pos=pos, color=255 * VIK + 0.5)

        self.color_bar_data["color_bar"].setColorMap(
            self.color_bar_data["cmap_dict"][self.color_bar_data["current_theme"]])
        self.color_bar_data["color_bar"].setLevels(values=tuple(levels))

    def pointClicked(self, scatter_item, spot, ev, point_item):
        if point_item in self.geo_col.selected_objects["points"]:
            return

        if self.drawing_object == "Points":
            return

        if self.point_text_item is not None:
            self.removeItem(self.point_text_item)
            self.point_text_item = None
        # point_item.hoverable = False
        # point_item.setScatterStyle("selected")
        # point_item.point.tree_item.setSelected(True)
        if self.drawing_object == "Bezier":
            self.geo_col.select_object(point_item.point)
            n_ctrl_pts = len(self.geo_col.selected_objects["points"])
            degree = n_ctrl_pts - 1
            msg = (f"Added control point to curve. Number of control points: "
                   f"{len(self.geo_col.selected_objects['points'])} "
                   f"(degree: {degree}). Press 'Enter' to generate the curve.")
            self.sigStatusBarUpdate.emit(msg, 0)
        elif self.drawing_object == "Beziers":
            self.geo_col.select_object(point_item.point)
            n_ctrl_pts = len(self.geo_col.selected_objects["points"])
            degree = n_ctrl_pts - 1
            msg = (f"Added control point to curve. Number of control points: "
                   f"{len(self.geo_col.selected_objects['points'])} "
                   f"(degree: {degree}). Press 'Enter' to generate the curve.")
            self.sigStatusBarUpdate.emit(msg, 0)
        elif self.drawing_object == "LineSegment":
            if len(self.geo_col.selected_objects["points"]) < 2:
                self.geo_col.select_object(point_item.point)
            if len(self.geo_col.selected_objects["points"]) == 1:
                self.sigStatusBarUpdate.emit("Next, choose the line's endpoint", 0)
            if len(self.geo_col.selected_objects["points"]) == 2:
                self.sigEnterPressed.emit()  # Complete the line after selecting the second point
        elif self.drawing_object == "LineSegments":
            if len(self.geo_col.selected_objects["points"]) < 2:
                self.geo_col.select_object(point_item.point)
            if len(self.geo_col.selected_objects["points"]) == 1:
                self.sigStatusBarUpdate.emit("Next, choose the line's endpoint", 0)
            if len(self.geo_col.selected_objects["points"]) == 2:
                point_sequence = PointSequence([pt for pt in self.geo_col.selected_objects["points"]])
                self.geo_col.add_line(point_sequence=point_sequence)
                self.clearSelectedObjects()
                self.sigStatusBarUpdate.emit("Choose the next line's start point", 0)
        elif self.drawing_object == "BezierAddPoint":
            if len(self.geo_col.selected_objects["points"]) < 2:
                self.geo_col.select_object(point_item.point)
            if len(self.geo_col.selected_objects["points"]) == 1:
                self.sigStatusBarUpdate.emit("Now, choose the preceding point in the sequence", 0)
            if len(self.geo_col.selected_objects["points"]) == 2:
                self.sigEnterPressed.emit()
        elif self.drawing_object == "Airfoil":
            self.geo_col.select_object(point_item.point)
            if len(self.geo_col.selected_objects["points"]) == 1:
                self.sigStatusBarUpdate.emit("Now, select the trailing edge point. For a blunt trailing edge, the "
                                             "point must have two associated lines (connecting to the upper and lower"
                                             " surface end points).", 0)
            elif len(self.geo_col.selected_objects["points"]) == 2:
                self.sigStatusBarUpdate.emit("Now, select the upper surface endpoint. For a sharp trailing edge, "
                                             "press the enter key to finish generating the airfoil.", 0)
            elif len(self.geo_col.selected_objects["points"]) == 3:
                self.sigStatusBarUpdate.emit("Now, select the lower surface endpoint.", 0)
            elif len(self.geo_col.selected_objects["points"]) == 4:
                self.sigEnterPressed.emit()
        elif self.drawing_object == "DistanceConstraint":
            self.geo_col.select_object(point_item.point)
            if len(self.geo_col.selected_objects["points"]) == 1:
                self.sigStatusBarUpdate.emit("Now, choose the last point", 0)
            elif len(self.geo_col.selected_objects["points"]) == 2:
                self.sigEnterPressed.emit()
        elif self.drawing_object in ["RelAngle3Constraint", "Perp3Constraint", "AntiParallel3Constraint"]:
            self.geo_col.select_object(point_item.point)
            if len(self.geo_col.selected_objects["points"]) == 1:
                self.sigStatusBarUpdate.emit("Now, choose the vertex", 0)
            elif len(self.geo_col.selected_objects["points"]) == 2:
                self.sigStatusBarUpdate.emit("Finally, choose the end point", 0)
            elif len(self.geo_col.selected_objects["points"]) == 3:
                self.sigEnterPressed.emit()
        elif self.drawing_object == "SymmetryConstraint":
            self.geo_col.select_object(point_item.point)
            if len(self.geo_col.selected_objects["points"]) == 1:
                self.sigStatusBarUpdate.emit("Now, choose the mirror axis end point", 0)
            elif len(self.geo_col.selected_objects["points"]) == 2:
                self.sigStatusBarUpdate.emit("Now, choose the tool point", 0)
            elif len(self.geo_col.selected_objects["points"]) == 3:
                self.sigStatusBarUpdate.emit("Finally, choose the target point", 0)
            elif len(self.geo_col.selected_objects["points"]) == 4:
                self.sigEnterPressed.emit()
        elif self.drawing_object == "ROCurvatureConstraint":
            self.geo_col.select_object(point_item.point)
            if len(self.geo_col.selected_objects["points"]) == 1:
                self.sigEnterPressed.emit()
        else:
            self.geo_col.select_object(point_item.point)

    def pointMoving(self, point: DraggablePoint):
        if self.point_text_item is not None:
            self.removeItem(self.point_text_item)
            self.point_text_item = None
        for curve in point.curveOwners:
            curve.updateCurveItem()

    @undoRedoAction
    def pointStartedMoving(self, point: DraggablePoint):
        pass

    def pointFinishedMoving(self, point: DraggablePoint):
        pass

    def setItemStyle(self, item, style: str):
        valid_styles = ["default", "hovered", "selected"]
        if style not in valid_styles:
            raise ValueError(f"Style found ({style}) is not a valid style. Must be one of {valid_styles}.")

        if style == "hovered":
            # Point
            if isinstance(item, DraggablePoint):
                self.point_hovered_item = item
                point = self.point_hovered_item
                if self.point_text_item is None:
                    self.point_text_item = pg.TextItem(
                        f"{point.point.name()}\nx: {point.point.x().value():.6f}\ny: {point.point.y().value():.6f}",
                        anchor=(0, 1))
                    self.point_text_item.setFont(QFont("DejaVu Sans", 8))
                    self.addItem(self.point_text_item)
                    self.point_text_item.setPos(point.point.x().value(), point.point.y().value())
                item.setScatterStyle(mode="hovered")
            # Curve
            elif isinstance(item, HoverableCurve):
                self.curve_hovered_item = item
                item.setCurveStyle("hovered")
            elif isinstance(item, ConstraintItem):
                self.constraint_hovered_item = item
                item.setStyle(mode="hovered")

        elif style == "default":
            # Point
            if isinstance(item, DraggablePoint):
                self.point_hovered_item = None
                self.removeItem(self.point_text_item)
                self.point_text_item = None
                item.setScatterStyle(mode="default")
            # Curve
            elif isinstance(item, HoverableCurve):
                self.curve_hovered_item = None
                item.setCurveStyle("default")
            elif isinstance(item, ConstraintItem):
                self.constraint_hovered_item = None
                item.setStyle(theme=self.gui_obj.themes[self.gui_obj.current_theme])

        elif style == "selected":
            # Point
            if isinstance(item, DraggablePoint):
                self.point_hovered_item = None
                self.removeItem(self.point_text_item)
                self.point_text_item = None
                item.setScatterStyle(mode="selected")
            elif isinstance(item, ConstraintItem):
                self.constraint_hovered_item = None
                item.setStyle(mode="selected")
            # Curve
            # elif isinstance(item, HoverableCurve):
            #     self.curve_hovered_item = None
            #     item.setCurveStyle("default")

    def pointHovered(self, scatter_item, spot, ev, point_item):
        if point_item.dragPoint is not None:
            return
        self.geo_col.hover_enter_obj(point_item.point)

    def pointLeaveHovered(self, scatter_item, spot, ev, point_item):
        self.geo_col.hover_leave_obj(point_item.point)

    def curveHovered(self, item):
        self.geo_col.hover_enter_obj(item.parametric_curve)

    def curveLeaveHovered(self, item):
        self.geo_col.hover_leave_obj(item.parametric_curve)

    def constraintHovered(self, item):
        self.geo_col.hover_enter_obj(item.canvas_item)

    def constraintLeaveHovered(self, item):
        self.geo_col.hover_leave_obj(item.canvas_item)

    def airfoil_hovered(self, airfoil: Airfoil, x_centroid: float, y_centroid: float):
        """
        Adds the name of the airfoil as a label to the airfoil's centroid when the airfoil shape is
        hovered with the mouse

        Parameters
        ==========
        airfoil_name: str
            Name of the airfoil (A0, A1, etc.)

        x_centroid: float
            x-location of the airfoil's centroid

        y_centroid: float
            y-location of the airfoil's centroid
        """

        # Get the color for the text
        # main_color = None if self.gui_obj is None else self.gui_obj.themes[self.gui_obj.current_theme]["main-color"]
        main_color = self.gui_obj.themes[self.gui_obj.current_theme]["main-color"]

        # Add the name of the airfoil as a pg.TextItem if not dragging
        text_item = pg.TextItem(airfoil.name(), anchor=(0.5, 0.5), color=main_color)
        text_item.setFont(QFont("DejaVu Sans", 10))
        self.airfoil_text_items[airfoil] = text_item
        self.airfoil_text_items[airfoil].setPos(x_centroid, y_centroid)
        self.addItem(self.airfoil_text_items[airfoil])

    def airfoil_exited(self, airfoil: Airfoil):
        """
        Remove the label from the airfoil centroid on mouse hover exit
        """
        if airfoil in self.airfoil_text_items.keys():
            self.removeItem(self.airfoil_text_items[airfoil])

    def removeCurve(self, item):
        self.geo_col.remove_pymead_obj(item.parametric_curve)

    def exportPlot(self):
        color_bar_data = self.color_bar_data
        current_min_level, current_max_level = None, None
        if color_bar_data is not None:
            current_levels = color_bar_data["color_bar"].levels()
            current_min_level = current_levels[0]
            current_max_level = current_levels[1]

        dialog = PlotExportDialog(self, gui_obj=self.gui_obj, theme=self.gui_obj.themes[self.gui_obj.current_theme],
                                  current_min_level=current_min_level,
                                  current_max_level=current_max_level)
        if dialog.exec_():
            # Get the inputs from the dialog
            inputs = dialog.value()

            # Create the pyqtgraph ImageExporter object from the airfoil canvas
            exporter = pg.exporters.ImageExporter(self.plot)

            # Set the image width parameter
            exporter.parameters()["width"] = inputs["width"]

            # Prevent saves to the local GUI directory
            if inputs["save_dir"] == "":
                self.gui_obj.disp_message_box(f"Please select a save directory")
                return

            # Prevent saves to an invalid file path
            if not os.path.exists(inputs["save_dir"]):
                self.gui_obj.disp_message_box(f"The save directory {inputs['save_dir']} does not exist",
                                              message_mode="error")
                return

            # If there is no .png extension, add one
            file_path = os.path.join(inputs["save_dir"], inputs["save_name"])
            if not os.path.splitext(file_path)[-1] == ".png":
                file_path += ".png"

            # Export the image
            exporter.export(file_path)

            # Display success message
            self.gui_obj.disp_message_box(f"Plot saved to {file_path}", message_mode="info")

    def contextMenuEvent(self, event):
        menu = QtWidgets.QMenu(self)
        menu.setStyleSheet(f"""
        QMenu::item:selected {{ background-color: 
            {self.gui_obj.themes[self.gui_obj.current_theme]['menu-item-selected-color']} }}
        """)
        create_geometry_menu = menu.addMenu("Create Geometry")
        modify_geometry_menu = menu.addMenu("Modify Geometry")
        add_constraint_menu = menu.addMenu("Add Constraint")

        removeCurveAction, curve_item, insertCurvePointAction, splitPolyAction = None, None, None, None
        if self.curve_hovered_item is not None:
            # Curve removal action
            removeCurveAction = modify_geometry_menu.addAction("Remove Curve")
            curve_item = self.curve_hovered_item

            # Curve point insertion action
            insertCurvePointAction = modify_geometry_menu.addAction("Insert Curve Point")

            if isinstance(curve_item.parametric_curve, PolyLine):
                splitPolyAction = modify_geometry_menu.addAction("Split PolyLine")

        drawPointAction = create_geometry_menu.addAction("Insert Point")
        drawBezierCurveThroughPointsAction = create_geometry_menu.addAction("Bezier Curve Through Points")
        drawLineSegmentThroughPointsAction = create_geometry_menu.addAction("Line Segment Through Points")
        generateAirfoilAction = create_geometry_menu.addAction("Generate Airfoil")
        generateMEAAction = create_geometry_menu.addAction("Generate MEA")
        addRelAngle3ConstraintAction = add_constraint_menu.addAction("Add Relative Angle 3 Constraint")
        addPerp3ConstraintAction = add_constraint_menu.addAction("Add Perpendicular 3 constraint")
        addAntiParallel3ConstraintAction = add_constraint_menu.addAction("Add Anti-Parallel 3 Constraint")
        addSymmetryConstraintAction = add_constraint_menu.addAction("Add Symmetry Constraint")
        addROCurvatureConstraintAction = add_constraint_menu.addAction("Add Radius of Curvature Constraint")
        addDistanceConstraintAction = add_constraint_menu.addAction("Add Distance Constraint")
        exportPlotAction = menu.addAction("Export Plot")
        view_pos = self.getPlotItem().getViewBox().mapSceneToView(event.pos())
        res = menu.exec_(event.globalPos())
        if res is None:
            return
        if res == drawPointAction:
            self.drawPoint(x=[view_pos.x()], y=[view_pos.y()])
        elif res == drawBezierCurveThroughPointsAction:
            self.drawBezier()
        elif res == drawLineSegmentThroughPointsAction:
            self.drawLineSegment()
        elif res == generateAirfoilAction:
            self.generateAirfoil()
        elif res == generateMEAAction:
            self.generateMEA()
        elif res == addRelAngle3ConstraintAction:
            self.addRelAngle3Constraint()
        elif res == addAntiParallel3ConstraintAction:
            self.addAntiParallel3Constraint()
        elif res == addPerp3ConstraintAction:
            self.addPerp3Constraint()
        elif res == addSymmetryConstraintAction:
            self.addSymmetryConstraint()
        elif res == addROCurvatureConstraintAction:
            self.addROCurvatureConstraint()
        elif res == addDistanceConstraintAction:
            self.addDistanceConstraint()
        elif res == removeCurveAction and curve_item is not None:
            self.removeCurve(curve_item)
        elif res == insertCurvePointAction and curve_item is not None:
            self.addPointToCurve(curve_item)
        elif res == splitPolyAction and curve_item is not None:
            self.splitPoly(curve_item)
        elif res == exportPlotAction:
            self.exportPlot()

    @undoRedoAction
    def removeSelectedPoints(self):
        self.geo_col.remove_selected_objects()

    def clearSelectedObjects(self):
        self.geo_col.clear_selected_objects()

    def onLineItemAdded(self, line_item):
        line_item.sigCurveHovered.connect(self.curveHovered)
        line_item.sigCurveNotHovered.connect(self.curveLeaveHovered)

    def arrowKeyPointMove(self, key, mods):
        step = 10 * self.geo_col.single_step if mods == Qt.ShiftModifier else self.geo_col.single_step
        if key == Qt.Key_Left:
            for point in self.geo_col.selected_objects["points"]:
                point.request_move(point.x().value() - step, point.y().value())
        elif key == Qt.Key_Right:
            for point in self.geo_col.selected_objects["points"]:
                point.request_move(point.x().value() + step, point.y().value())
        elif key == Qt.Key_Up:
            for point in self.geo_col.selected_objects["points"]:
                point.request_move(point.x().value(), point.y().value() + step)
        elif key == Qt.Key_Down:
            for point in self.geo_col.selected_objects["points"]:
                point.request_move(point.x().value(), point.y().value() - step)

    def mousePressEvent(self, ev):

        super().mousePressEvent(ev)

        if not self.drawing_object == "Points":
            return

        view_pos = self.getPlotItem().getViewBox().mapSceneToView(ev.pos())
        self.geo_col.add_point(view_pos.x(), view_pos.y())

    @undoRedoAction
    def removeSelectedObjects(self):
        self.geo_col.remove_selected_objects()

    def canvasShortcuts(self):
        return {
            Qt.Key_P: self.drawPoints,
            Qt.Key_L: self.drawLines,
            Qt.Key_B: self.drawBeziers,
            Qt.Key_F: self.generateAirfoil,
            Qt.Key_W: self.generateWebAirfoil,
            Qt.Key_M: self.generateMEA,
            Qt.Key_D: self.addDistanceConstraint,
            Qt.Key_A: self.addRelAngle3Constraint,
            Qt.Key_T: self.addPerp3Constraint,
            Qt.Key_H: self.addAntiParallel3Constraint,
            Qt.Key_S: self.addSymmetryConstraint,
            Qt.Key_R: self.addROCurvatureConstraint
        }

    def keyPressEvent(self, ev):
        key = ev.key()
        mods = QApplication.keyboardModifiers()
        if key == Qt.Key_Return:
            self.sigEnterPressed.emit()
        elif key == Qt.Key_Escape:
            self.sigEscapePressed.emit()
            self.geo_col.clear_selected_objects()
            self.sigStatusBarUpdate.emit("", 0)
        elif key == Qt.Key_Delete:
            self.removeSelectedObjects()
            self.sigStatusBarUpdate.emit("", 0)
        elif key in (Qt.Key_Left, Qt.Key_Right, Qt.Key_Down, Qt.Key_Up) and len(self.geo_col.selected_objects["points"]) > 0:
            self.arrowKeyPointMove(ev.key(), mods)
        elif key in self.canvasShortcuts():
            self.canvasShortcuts()[key]()
